import os
os.environ["MKL_NUM_THREADS"] = "1" 
os.environ["NUMEXPR_NUM_THREADS"] = "1" 
os.environ["OMP_NUM_THREADS"] = "1" 
os.environ['OPENBLAS_NUM_THREADS'] = "1"


import numpy as np
import scipy
import logging
import traceback
from xframe import log
os.chdir(os.path.expanduser('~/Programs/xframe'))
from xframe.startup_routines import load_recipes
from xframe.startup_routines import dependency_injection_no_soft
from xframe.library import mathLibrary
from xframe.externalLibraries.flt_plugin import LegendreTransform
mathLibrary.leg_trf = LegendreTransform
from xframe.externalLibraries.shtns_plugin import sh
mathLibrary.shtns = sh
from xframe.control.Control import Controller
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import cross_correlation_to_deg2_invariant
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import ccd_remove_0_order
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import cross_correlation_mask
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_apply_precision_filter
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import ccd_to_deg2_invariant_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import bl_3d_pseudo_inverse_worker
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_cc_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import density_to_deg2_invariants_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_projection_matrices
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_projection_matrices_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_eigenvalues
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import ccd_associated_legendre_matrices
from xframe.plugins.MTIP.analysisLibrary import fxs_invariant_tools as i_tools
from xframe.plugins.MTIP.analysisLibrary.hankel_transforms import generate_weightDict_zernike_spherical
from xframe.plugins.MTIP.analysisLibrary.harmonic_transforms import HarmonicTransform
from xframe.plugins.MTIP.analysisLibrary.fourier_transforms import generate_zernike_spherical_ft
from xframe.plugins.MTIP.analysisLibrary.ft_grid_pairs import get_grid
from xframe.plugins.MTIP.analysisLibrary.fxs_Projections import ReciprocalProjection_old,ReciprocalProjection
from xframe import settings
from xframe.library import mathLibrary as mLib
from xframe import Multiprocessing
from xframe.library import physicsLibrary as pLib
from xframe.library.gridLibrary import GridFactory
from xframe.presenters.matplolibPresenter import heat2D,heat2D_multi,plot1D
from xframe import Multiprocessing
from scipy.interpolate import interp1d
from scipy.linalg import solve_triangular

log=log.setup_custom_logger('root','INFO')

dependency_injection_no_soft()
from xframe.library.mathLibrary import gsl
#analysis_sketch = load_recipes(ana_name='MTIP.reconstruct', ana_settings='3d_model_0')[0]
analysis_sketch = load_recipes(ana_name='MTIP.reconstruct', ana_settings='3d_pentagon_10p_spher')[0] 
db = analysis_sketch[1]
opt = settings.analysis
controller = Controller(analysis_sketch)


ccd = db.load('ccd')
ccd_1 = db.load_ccd('/gpfs/exfel/theory_group/user/berberic/MTIP/ccd/3d_model_0.h5')
#ccd_1 = db.load_ccd('/gpfs/exfel/theory_group/user/berberic/MTIP/ccd/3d_pentagon_10particles_rectSim.h5')
cc = ccd['cross_correlation']
cc_1 = ccd_1['cross_correlation']

pre_init_opt = settings.analysis.projections.reciprocal.pre_init
proj_opt = pre_init_opt.cross_correlation

def binned_mean(cc_mask,cc,max_order,phis):
    # create bin ids
    step_size = np.pi/max_order
    n_bins = 2*max_order
    ids = (phis+step_size/2)//step_size
    n_roll= np.sum(ids == n_bins)
    ids[ids == n_bins] = 0

    # roll such that bin ids monotonically increase
    ccr = np.roll(np.array(cc),n_roll,axis = -1)
    cc_mask_r = np.roll(cc_mask,n_roll,axis = -1)
    idr = np.roll(ids,n_roll)

    # masked add
    ccr[~cc_mask_r] = 0.0
    split_ids = np.where(np.roll(idr,1)!=idr)[0]
    new_cc = np.add.reduceat(ccr,split_ids,axis = -1)

    # divide by number of unmasked values per bin
    n_unmasked_values = np.add.reduceat(cc_mask_r,split_ids,axis = -1)
    new_cc_mask = (n_unmasked_values!=0)
    new_cc[new_cc_mask]/=n_unmasked_values[new_cc_mask]
    new_phis = np.arange(n_bins)*2*np.pi/n_bins
    
    return new_cc,new_cc_mask,new_phis

#metadata = {**ccd,**proj_opt}  
#cc_mask = cross_correlation_mask(ccd['data_grid'],metadata)
#locals().update(binned_mean(cc_mask,cc,99,ccd['data_grid']['phis']))
    

def interpolate(cc,mask,phis):
    shape = cc.shape
    n_phi = len(phis)
    n_pi_half=n_phi//4
    points = np.arange(n_phi)
    cci = np.roll(cc,n_pi_half,axis = -1)
    cc_old = np.array(cci) 
    mask = np.roll(mask,n_pi_half,axis = -1)
    cci = cci.reshape(-1,n_phi)
    mask = mask.reshape(-1,n_phi)
    for d,m in zip(cci,mask):
        #log.info('data shape {} mask shape {}'.format(d.shape,m.shape))
        unmasked_d = d[m]
        if len(unmasked_d)!=0:
            d[~m] = interp1d(phis[m],d[m])(phis[~m])
    cci = cci.reshape(shape)
    mask = mask.reshape(shape)
    cci = np.roll(cci,-n_pi_half,axis = -1)
    mask = np.roll(mask,-n_pi_half,axis = -1)
    return cci
    
#locals().update(interpolate(new_cc,new_cc_mask,new_phis))

def init(): 
    thetas = ccd['data_grid']['thetas']
    phis = ccd['data_grid']['phis']
    max_order = 99
    orders = np.arange(max_order+1)
    
    metadata = {**ccd,**proj_opt,'orders':np.arange(max_order+1),'extract_odd_orders':False,'bl_enforce_psd':False}          
    cc_mask = cross_correlation_mask(ccd['data_grid'],metadata)
    #cc_mask = np.full(cc.shape,True)
    new_cc_1,new_cc_mask_1,new_phis_1 = binned_mean(np.full(cc_mask.shape,True),cc_1,max_order,phis)
    new_cc_1 = interpolate(new_cc_1,new_cc_mask_1,new_phis_1)
    cns_1 = mLib.circularHarmonicTransform_real_forward(new_cc_1)
    new_cc,new_cc_mask,new_phis = binned_mean(cc_mask,cc,max_order,phis)
    new_cc = interpolate(new_cc,new_cc_mask,new_phis)
    cns = mLib.circularHarmonicTransform_real_forward(new_cc)
    #pl,ls,ms = mLib.gsl.legendre_sphPlm_array(max_order,max_m,np.cos(thetas),return_orders = True,sorted_by_l = True)
    qq_matrix = ccd_associated_legendre_matrices(ccd['data_grid']['thetas'],max_order,max_order)
    #qq_matrix = q_matrix[None,:]*q_matrix[:,None]/(2*orders+1)[None,None,None,:]
    inverses = np.linalg.inv(qq_matrix[...,2::2,2::2])
    q_ids = np.arange(len(thetas))
    
    bls = np.zeros_like(cns)
    bls[...,2::2] = np.sum(inverses*cns[...,None,2:max_order+1:2],axis =-1)
    bls[...,1:]/=metadata['n_particles']
    bls[...,0]/=metadata['n_particles']**2
    #bls_pseudo = np.swapaxes(cross_correlation_to_deg2_invariant(cc,3,**metadata)[0],0,-1)
    #Ms = qq_matrix[...,2::2,2::2]
    #Cs = cns[...,2:max_order+1:2]
    #bls[...,2::2] = np.array(tuple(solve_triangular(m,v) for m,v in zip(Ms.reshape(-1,*Ms.shape[-2:]),Cs.reshape(-1,*Cs.shape[-1:])) )).reshape(*Cs.shape)
    #bls_t = np.array(tuple(mLib.optimal_tikhonov_regularization(qq_matrix[i],cns[i],allow_offset=True)[0] for i in range(len(bls))))
    #bls_t = Multiprocessing.comm_module.request_mp_evaluation(mLib.optimal_tikhonov_regularization_worker,argArrays = [q_ids,q_ids],const_args=[qq_matrix,cns[...,:max_order+1].astype(complex),True],callWithMultipleArguments=True)
    cns_new = np.zeros_like(cns)
    #cns_new[...,2::2] = np.sum(qq_matrix[...,2::2,2::2]*bls[...,None,2:max_order+1:2],axis =-1)
    #cns/=10
    return locals()

locals().update(init())

def plot():
    qs = ccd['data_grid']['qs']
    qq_grid = GridFactory.construct_grid('uniform',[qs,qs])
    orders = np.arange(2,12,2)
    orders2 = np.arange(12,22,2)
    bls_plot = np.swapaxes(bls,0,-1)#/100**2
    bls_plot[0]#/=100
    cns_plot = np.swapaxes(cns,0,-1)#/100**2
    cns_plot_1 = np.swapaxes(cns_1,0,-1)#/100**2
    cns_new_plot = np.swapaxes(cns_new,0,-1)#/100**2
    cns_diff = np.abs(cns_plot - cns_new_plot)/np.abs(cns_plot)
    cns_diff_1_100 = np.abs(cns_plot_1-cns_plot)/np.abs(cns_plot_1)
    bls_plot_psd = np.zeros_like(bls_plot)
    q_start = 25
    bls_plot_psd[:,q_start:,q_start:] = mLib.nearest_positive_semidefinite_matrix(bls_plot[:,q_start:,q_start:])
    layout = {'title':'','x_label':'$q_1$', 'y_label':'$q_2$','title_size':40,'text_size':20}#,))*len(orders)
    #bls_plot[:,30:,30:] = mLib.nearest_positive_semidefinite_matrix(bls_plot[:,30:,30:])
    fig = heat2D_multi.get_fig((
        list(np.abs(cns_plot_1[orders])),
        list(np.abs(bls_plot[orders])),
        list(np.abs(bls_plot_psd[orders]))
    ),grid = qq_grid,shape = (3,len(orders)),size=(5*len(orders)+2*5,3*5),vmin = 1e+14, vmax = 1e+24,scale='log',cmap='plasma',layout = layout)
    #db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/bl_from_cm/bl0-10_10p_mO70_rect.matplotlib',fig)
    db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/bl_from_cm/bl0-20_10p_mO99_rect.matplotlib',fig)
    #fig = heat2D_multi.get_fig((
    #    list(np.abs(cns_diff_1_100[orders])[:,1:,1:])
    #    ,),grid = qq_grid[1:,1:],shape = (1,len(orders)),size=(5*len(orders)+2*5,1*5),cmap='plasma')
    fig = heat2D_multi.get_fig((list(np.abs(cns_diff_1_100[orders])),list(np.abs(cns_diff_1_100[orders2]))),grid = qq_grid,shape = (2,len(orders)),size=(5*len(orders)+2*5,2*5),vmin = 1e-2, vmax = 1e2,scale='log',cmap='plasma',layout = layout)
    #db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/bl_from_cm/cn0-10_p1_p10_mO30_diff_rect.matplotlib',fig)
    #db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/bl_from_cm/cn0-20_p1_p10_mO99_diff_spher.matplotlib',fig)
    #fig = heat2D_multi.get_fig((
    #    list(np.abs(cns_diff[orders]))
    #),grid = qq_grid,shape = (1,len(orders)),size=(5*len(orders)+2*5,1*5),scale='log',cmap='plasma')    

plot()



def init_transforms():
    cc_data = ccd_1['cross_correlation'][...,:800]
    maxQ = np.max(ccd_1['radial_points'])*2
    max_order = 99
    n_radial_points = 128#300
    opt=settings.analysis
    ht_opt={'dimensions':3,'max_order':max_order,'indices':'lm','anti_aliazing_degree':2,'n_phi':False,'n_theta':False}
    #ht_opt={'dimensions':opt.dimensions,**opt['grid'],'type':opt['fourier_transform']['type'],'pi_in_q':opt['fourier_transform']['pi_in_q'],'reciprocal_cut_off':maxQ}
    # harmonic transforms and grid
    cht=HarmonicTransform('complex',ht_opt)
    weight_dict = generate_weightDict_zernike_spherical(max_order,n_radial_points,expansion_limit = 2000)
    grid_opt={'dimensions':3,'type':'Zernike_dict','max_q':maxQ,'n_radial_points':n_radial_points,**cht.grid_param,'pi_in_q':True}
    grid_pair = get_grid(grid_opt)
    maxR = grid_pair.real[:,0,0].max()
    ft,ift = generate_zernike_spherical_ft(maxR,weight_dict,cht,'complex',True,use_gpu=True)    
    return locals()

#locals().update(init_transforms())

def generate_density_spheres(grid,radius = 100, norm = 'standard',cart_centers= np.array([[0,200,0],[0,-200,0],[200,0,0],[-200,0,0],[400,0,0],[0,0,300]],dtype = np.float64),random_orientation = False):
    centers = mLib.cartesian_to_spherical(cart_centers)
    #centers = [(100,0,np.pi),(100,np.pi,np.pi),(100,0,0),(200,np.pi,0),(300,np.pi,np.pi/2)]
    sphere_funcs = [mLib.SampleShapeFunctions.get_disk_function(radius,center = center,norm=norm,random_orientation = random_orientation) for center in centers]
    density = sphere_funcs.pop(0)(grid)
    for sphere_func in sphere_funcs:
        density += sphere_func(grid)
    return density


def init2():
    r = 65
    phis = np.arange(5)*2*np.pi/5
    centers = 2*r*np.stack((np.sin(phis),np.cos(phis),np.zeros(5)),axis = -1)
    centers = list(centers)
    centers.append(np.array([0.0,0.0,0.0]))
    centers = np.array(centers)
    centers2 = np.array([centers[0],centers[2],centers[3]])
    density1 = generate_density_spheres(grid_pair.real,radius = r,cart_centers=centers)
    density2 = generate_density_spheres(grid_pair.real,radius = r,cart_centers=centers2)
    density = 25*(density1+density2)
    #bls_exact = density_to_deg2_invariants_3d(density,cht,ft)
    
    return locals()

#locals().update(init2())

def plot2():
    orders = np.arange(0,10,2)
    #orders = np.arange(10,100,10)
    #orders = [2]
    qs = grid_pair.reciprocal[:,0,0,0]
    qq_grid = GridFactory.construct_grid('uniform',[qs,qs])
    bls = bls_exact
    
    fig = heat2D_multi.get_fig(list(np.abs(bls[orders])),grid = qq_grid,shape = (1,len(orders)),size=(5*len(orders)+2*5,5),vmin = 1e+14, vmax = 1e+26,scale='log',cmap='plasma')
    db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/bl_from_cm/bl_exact0-10.matplotlib',fig)

    fig = heat2D_multi.get_fig(list(np.abs(mLib.nearest_positive_semidefinite_matrix(bls[orders]))),grid = qq_grid,shape = (1,len(orders)),size=(5*len(orders)+2*5,5),vmin = 1e+14, vmax = 1e+26,scale='log',cmap='plasma')
    db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/bl_from_cm/bl_exact_psd0-10.matplotlib',fig)

    intensity_exact = (np.abs(ft(density))**2).real
    Ilm = cht.forward(intensity_exact)
    fig = plot1D.get_fig(np.abs(Ilm[2][:,0]),y_scale='log',grid = qs)
    db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/bl_from_cm/Ilm_exact.matplotlib',fig)
    #bl2_new = mLib.nearest_positive_semidefinite_matrix(bls[...,2])
    #fig = heat2D.get_fig(np.abs(bl2_new.real),scale='log')
    #db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/bl_from_cm/bl_exact2n.matplotlib',fig)
    #bl2_new = (bls[...,2]+bls[...,2].T)/2
    #fig = heat2D.get_fig(np.abs(bl2_new.real),scale='log')
    #db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/bl_from_cm/bl_exact2nn.matplotlib',fig)
    #
    #s,v = np.linalg.eigh(bl2_new)
    #s[s<s.max()*1e-15]=0
    #bl2_new2 = v*s[None,:] @ v.T 
    #fig = heat2D.get_fig(np.abs(bl2_new2.real),scale='log')
    #db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/bl_from_cm/bl_exact2nnn.matplotlib',fig)

#plot2()



#rd = ft(density)
#I = np.abs(rd*rd.conj()).real
#Ilm = cht.forward(I)
#phases = rd/np.sqrt(I)
#
#multiplicies = [1,0,1/100,1/10,1/2,2,10,100]
#intensities = []
#for m in multiplicies:
#    temp = [np.copy(Il)*m for Il in Ilm]
#    temp[0] = np.copy(Ilm[0])
#    intensities.append(cht.inverse(temp))
#
#db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/bl_from_cm/N_particles_shift.vtk',[d.copy() for d in intensities],grid=grid_pair.reciprocal,grid_type='spherical')




import numpy as np
import logging

log=logging.getLogger('root')
from xframe.library import pythonLibrary as pyLib
xprint = pyLib.xprint
from xframe.interfaces import CommunicationInterface as ComInterfaceAnalysis
# in place for future upgrade to MPI
from xframe import Multiprocessing
from xframe.Multiprocessing import MPMode,MPMode_Queue
cpu_count = Multiprocessing.get_local_cpu_count


class SingleProcessCommunictionAnalysis(ComInterfaceAnalysis):
    def __init__(self,control_worker=False):
        self.control_worker=control_worker

    def get_data(self,opt):
        data=self.control_worker.processDataRequest_SingleProcess(opt)
        return data

    def get_experiment(self):
        return self.control_worker.experiment_worker
    
    def get_geometry(self,approximation='None',out_coord_sys='spherical'):
        geometry=self.control_worker.process_geometry_request_single_thread(approximation=approximation,out_coord_sys = out_coord_sys)
        return geometry
    def get_exp_info(self):
        return self.control_worker.process_exp_info_request_single_thread()

    def request_mp_evaluation(self,
                                func,
                                mode:MPMode = MPMode_Queue(),
                                is_daemon=False,
                                return_process_objects = False,
                                input_arrays=[],
                                const_inputs=[],
                                call_with_multiple_arguments=False,
                                split_mode='sequential',
                                split_together=False,
                                n_processes = False,
                                event_names=[],
                                create_queues = False):
        result = self.control_worker.process_mp_request_single_node(func,
                                 mode = mode,
                                is_daemon=is_daemon,
                                return_process_objects = return_process_objects,
                                input_arrays=input_arrays,
                                const_inputs=const_inputs,
                                call_with_multiple_arguments=call_with_multiple_arguments,
                                split_mode=split_mode,
                                split_together=split_together,
                                n_processes = n_processes,
                                event_names=event_names,
                                create_queues = create_queues)
        return result
    def get_mp_lock(self):
        return self.control_worker.mp_lock

    def n_cpus(self):
        return cpu_count()

    def free_mem(self):
        return Multiprocessing.get_free_memory()

    def getGatheredResults():
        pass
    
    def sendResults():
        pass

    def put_string_to_controller(self,string):
        cpu_id = Multiprocessing.get_process_id()
        log.info('cpu id = {}'.format(cpu_id))
        self.control_worker.queue.put([cpu_id,string])
        Multiprocessing.wait_for_cpu_event(cpu_id)
        Multiprocessing.clear_cpu_event(cpu_id)        

    def add_gpu_process(self,gpu_process):
        control_worker = self.control_worker.select_control_worker()
        handle = self.control_worker.gpu_manager.request_gpu_handle_client(control_worker.name,gpu_process)
        return handle.run
    def restart_control_worker(self):
        self.control_worker.restart_working()


class MPICommunictionAnalysis(ComInterfaceAnalysis):
    def __init__(self):
        pass
    
    def getData(self):
        pass

    def getGeometry(self):
        pass

class MPICommunictionExperiment(ComInterfaceAnalysis):
    def __init__(self):        
        pass
    
    def getData(self):
        pass

    def getGeometry(self):
        pass



import os
os.environ["MKL_NUM_THREADS"] = "1" 
os.environ["NUMEXPR_NUM_THREADS"] = "1" 
os.environ["OMP_NUM_THREADS"] = "1" 
os.environ['OPENBLAS_NUM_THREADS'] = "1"


import numpy as np
import scipy
import logging
import traceback
from xframe import log
os.chdir(os.path.expanduser('~/Programs/xframe'))
from xframe.startup_routines import load_recipes
from xframe.startup_routines import dependency_injection_no_soft
from xframe.library import mathLibrary
from xframe.externalLibraries.flt_plugin import LegendreTransform
mathLibrary.leg_trf = LegendreTransform
from xframe.externalLibraries.shtns_plugin import sh
mathLibrary.shtns = sh
from xframe.control.Control import Controller
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import cross_correlation_to_deg2_invariant
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import ccd_remove_0_order,cross_correlation_low_pass
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import cross_correlation_mask
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_apply_precision_filter
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import ccd_to_deg2_invariant_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import bl_3d_pseudo_inverse_worker
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_cc_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import density_to_deg2_invariants_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_projection_matrices
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_projection_matrices_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_eigenvalues
from xframe.plugins.MTIP.analysisLibrary import fxs_invariant_tools as i_tools
from xframe.plugins.MTIP.analysisLibrary.hankel_transforms import generate_weightDict_zernike_spherical
from xframe.plugins.MTIP.analysisLibrary.harmonic_transforms import HarmonicTransform
from xframe.plugins.MTIP.analysisLibrary.fourier_transforms import generate_zernike_spherical_ft
from xframe.plugins.MTIP.analysisLibrary.ft_grid_pairs import get_grid
from xframe.plugins.MTIP.analysisLibrary.fxs_Projections import ReciprocalProjection_old,ReciprocalProjection
from xframe import settings
from xframe.library import mathLibrary as mLib
from xframe import Multiprocessing
from xframe.library import physicsLibrary as pLib
from xframe.library.gridLibrary import GridFactory
from xframe.library.mathLibrary import nearest_positive_semidefinite_matrix
from scipy import signal

log=log.setup_custom_logger('root','INFO')

dependency_injection_no_soft()
#analysis_sketch = load_recipes(ana_name='MTIP.reconstruct', ana_settings='3d_model_0')[0]
analysis_sketch = load_recipes(ana_name='MTIP.reconstruct', ana_settings='3d_pentagon_100')[0] 
db = analysis_sketch[1]
opt = settings.analysis
controller = Controller(analysis_sketch)
from xframe.presenters.matplolibPresenter import heat2D,heat2D_multi

def init_transforms():
    ccd= db.load('ccd')
    cc_data = ccd['cross_correlation'][...,:800]
    maxQ = np.max(ccd['radial_points'])
    max_order = 99
    n_radial_points = 128
    opt=settings.analysis
    ht_opt={'dimensions':3,'max_order':max_order,'indices':'lm','anti_aliazing_degree':2,'n_phi':0,'n_theta':0}
    #ht_opt={'dimensions':opt.dimensions,**opt['grid'],'type':opt['fourier_transform']['type'],'pi_in_q':opt['fourier_transform']['pi_in_q'],'reciprocal_cut_off':maxQ}
    # harmonic transforms and grid
    cht=HarmonicTransform('complex',ht_opt)
    weight_dict = generate_weightDict_zernike_spherical(max_order,n_radial_points)
    grid_opt={'dimensions':3,'type':'Zernike_dict','max_q':maxQ,'n_radial_points':n_radial_points,**cht.grid_param,'pi_in_q':True}
    grid_pair = get_grid(grid_opt)
    maxR = grid_pair.real[:,0,0].max()
    ft,ift = generate_zernike_spherical_ft(maxR,weight_dict,cht,'complex',True,use_gpu=True)
    return locals()


def generate_density_spheres(grid,radius = 100, norm = 'standard',cart_centers= np.array([[0,200,0],[0,-200,0],[200,0,0],[-200,0,0],[400,0,0],[0,0,300]],dtype = np.float64),random_orientation = False):
    centers = mLib.cartesian_to_spherical(cart_centers)
    #centers = [(100,0,np.pi),(100,np.pi,np.pi),(100,0,0),(200,np.pi,0),(300,np.pi,np.pi/2)]
    sphere_funcs = [mLib.SampleShapeFunctions.get_disk_function(radius,center = center,norm=norm,random_orientation = random_orientation) for center in centers]
    density = sphere_funcs.pop(0)(grid)
    for sphere_func in sphere_funcs:
        density += sphere_func(grid)
    return density


def init():
    ccd = db.load('ccd')
    ccd_1 = db.load_ccd('/gpfs/exfel/theory_group/user/berberic/MTIP/ccd/3d_model_0.h5')
    #sys._getframe(1).f_locals.update(init_transforms())
    #density = generate_density_spheres(grid_pair.real,radius = 200,norm='inf',cart_centers = np.array([[-200,0,0],[200,0,0]],dtype=np.float64),random_orientation = True)
    #density += generate_density_spheres(grid_pair.real,radius = 100,norm='inf',cart_centers = np.array([[0,300,0],[0,-100,0]],dtype=np.float64),random_orientation = False)
    #Bl = density_to_deg2_invariants_3d(density,cht,ft).real
    cc = ccd['cross_correlation']
    cc_1 = ccd_1['cross_correlation']
    grid = GridFactory.construct_grid('uniform',[ccd['qs'],ccd['qs']])

    max_order = 99
    proj_opt = settings.analysis.projections.reciprocal.pre_init.cross_correlation
    metadata = {**ccd,'orders':np.arange(max_order+1),'mask':proj_opt.mask,'n_particles':1,'extract_odd_orders':False}
    #cc_1,_locals = cross_correlation_low_pass(cc_1,max_order)
    bl_single = cross_correlation_to_deg2_invariant(cc_1,3,metadata)[0]
    
    #new_cc,_locals = cross_correlation_low_pass(cc,max_order)
    new_cc,_locals = cross_correlation_low_pass(cc,99)
    #sos_cc = signal.butter(1, 700, 'lp', fs=1600, output='sos')
    #new_cc = signal.sosfilt(sos_cc,cc,axis = -1)
    #new_cc = cc
    
    cc_3 = np.copy(new_cc)
    low_pass_order = 30
    sos = signal.butter(1, low_pass_order, 'lp', fs=128, output='sos')
    new_cc = signal.sosfilt(sos,new_cc,axis = 0)
    new_cc = signal.sosfilt(sos,new_cc,axis = 1)
    #new_cc = cc
    
    metadata = {**ccd,'orders':np.arange(max_order+1),'mask':proj_opt.mask,'n_particles':proj_opt.n_particles,'extract_odd_orders':False}
    metadata1 = {**ccd,'orders':np.arange(max_order+1),'mask':proj_opt.mask,'n_particles':1,'extract_odd_orders':False}
    bl_new = cross_correlation_to_deg2_invariant(new_cc,3,metadata)[0]
    bl3 = cross_correlation_to_deg2_invariant(cc_3,3,metadata)[0]

    
    A=nearest_positive_semidefinite_matrix(bl_new)
    AA=nearest_positive_semidefinite_matrix(bl3)
    A2 = signal.sosfilt(sos, AA,axis = -1)
    A2 = signal.sosfilt(sos, A2,axis = -2)
    bl_low_pass = signal.sosfilt(sos, bl3,axis = -1)
    bl_low_pass = signal.sosfilt(sos, bl3,axis = -2) 
    A3 = nearest_positive_semidefinite_matrix(bl_low_pass)
    #A4 = signal.sosfilt(sos, bl_single,axis = -1)
    #A4 = signal.sosfilt(sos, A4,axis = -2)
    A4 = nearest_positive_semidefinite_matrix(bl_single)

    

    #A2 = nearest_positive_semidefinite_matrix(
    #    cross_correlation_to_deg2_invariant(
    #        deg2_invariant_to_cc_3d(A,ccd['xray_wavelength'],ccd['data_grid'],np.arange(100))
    #        ,3,metadata1)[0]
    #)
    return locals()

#locals().update(init())

def plot(orders):
    b1 = tuple(np.abs(bl_single[orders]))
    b100 = tuple(np.abs(bl_new[orders]))
    bA = tuple(np.abs(A[orders]))
    bA2 = tuple(np.abs(A2[orders]))
    bA3 = tuple(np.abs(A3[orders]))
    bA4 = tuple(np.abs(A4[orders]))
    datasets = [b1,b100,bA,bA2,bA3,bA4]

    n_orders = len(orders)
    
    layout_1 = {'title':'S','x_label':'$q_1$', 'y_label':'$q_2$','title_size':40,'text_size':30}
    layout_100 = {'title':'M l-cc-phi-q','x_label':'$q_1$', 'y_label':'$q_2$','title_size':40,'text_size':30}
    layout_A = {'title':'M l-cc-phi-q psd','x_label':'$q_1$', 'y_label':'$q_2$','title_size':40,'text_size':30}
    layout_A2 = {'title':'M l-cc-phi psd l-bl-q','x_label':'$q_1$', 'y_label':'$q_2$','title_size':40,'text_size':30}
    layout_A3 = {'title':'M  l-cc-phi l-bl-q psd','x_label':'$q_1$', 'y_label':'$q_2$','title_size':40,'text_size':30}
    layout_A4 = {'title':'S psd','x_label':'$q_1$', 'y_label':'$q_2$','title_size':40,'text_size':30}
    layout = tuple(zip(*[[layout_1]*(n_orders+1),
                         [layout_100]*(n_orders+1),
                         [layout_A]*(n_orders+1),
                         [layout_A2]*(n_orders+1),
                         [layout_A3]*(n_orders+1),
                         [layout_A4]*(n_orders+1),
                         ]))

    db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/cc_remove_zero_component/b0_comparison_psd.matplotlib',heat2D_multi.get_fig(tuple(zip(*datasets)),layout=layout, grid = grid,scale = 'log',shape = [n_orders,len(datasets)],size = (10*(len(datasets)+1),10*(n_orders)),vmin= 1e14, vmax = 1e24,cmap = 'plasma'))

grid = GridFactory.construct_grid('uniform',[ccd['qs'],ccd['qs']])    
orders = [2,4,6,8]
#plot(orders)

A = np.random.rand(4,10,10)
AA = A[2]

B = (A + np.swapaxes(A,-1,-2)) / 2
l,v = np.linalg.eigh(B)
l[l<0]=0
A2 =v*l[...,None,:] @ np.swapaxes(v,-1,-2)

B2 = (AA + AA.T) / 2
l2,v2 = np.linalg.eigh(B2)
l2[l2<0]=0
A3 =v2 @ np.diag(l2) @ v2.T




import numpy as np
import os
os.chdir('/gpfs/exfel/theory_group/user/berberic/Programs/xframe')
from xframe.database.euxfel import default_DB
from xframe.library.gridLibrary import GridFactory
from xframe.startup_routines import dependency_injection
dependency_injection()
from xframe.presenters.matplolibPresenter import plot1D,heat2D_multi,heat2D,heatPolar2D,agipd_heat
from xframe.library.mathLibrary import circularHarmonicTransform_real_forward,cartesian_to_spherical,spherical_to_cartesian
from xframe.library.mathLibrary import circularHarmonicTransform_complex_forward

from xframe.plugins.fxs3046_online.analysisLibrary.regrid2 import NearestNeighbourAgipdRegridder,create_lookup_array_2D,NearestNeighbourRegridder

from scipy.spatial.kdtree import KDTree
from scipy.ndimage import gaussian_filter

db = default_DB()

from xframe.startup_routines import load_recipes
from xframe.control.Control import Controller
analysis_sketch,exp_sketch = load_recipes(ana_name='fxs3046_online.generate_plots_2', ana_settings='hist_mean_162',exp_name='fxs3046_online',exp_settings = 'proc_-20_20_03_4')
controller = Controller(analysis_sketch,experiment_sketch = exp_sketch)
controller.chooseJobSingleProcess()
from xframe import database
from xframe import Multiprocessing

comm = Multiprocessing.comm_module
g = comm.get_geometry()

base_path = '/gpfs/exfel/theory_group/user/berberic/fxs3046/test/regrid2/'

pixel_centers = g['framed_pixel_centers']
data = np.zeros((16,512,128))
for module in data:
    for n,i in enumerate(module):
        i[n%2::2]=1

grid_cart = spherical_to_cartesian(g['pixel_grid'])
mask = g['mask']
fmask = g['framed_mask']
#fig=agipd_heat.get_fig(data,grid_cart,mask)
#db.save(base_path + 'data.matplotlib',fig)

#xs = np.arange(-5.5,6,1)
#pixel_centers = GridFactory.construct_grid('uniform',[xs,xs])[:]
#pixel_centers_polar = cartesian_to_spherical(pixel_centers)
#data = np.zeros(pixel_centers.shape[:-1])
#for n,i in enumerate(data[1:-1]):
#    i[1+n%2:-1:2]=1


#fig = heat2D.get_fig(data,grid = pixel_centers)

bound = np.min(np.max(spherical_to_cartesian(pixel_centers),axis = (0,1,2))[:2])
rs = np.arange(1024)*bound/1024
phis = np.arange(0,3200)*2*np.pi/3200
polar_grid = GridFactory.construct_grid('uniform',[rs,phis])[:]
cart_grid = spherical_to_cartesian(polar_grid)
#pixel_size = 2
#
rg = NearestNeighbourRegridder(pixel_centers,polar_grid)
#
framed_data = np.zeros(fmask.shape,dtype = float)
framed_data[fmask]=data.flatten()
polar_data = rg.apply(framed_data)

db.save(base_path + 'polar_data.vts',[polar_data],grid = polar_grid,grid_type='polar')


#fig_polar = heatPolar2D.get_fig(polar_data,grid = np.swapaxes(polar_grid,0,1))
#db.save(base_path + 'polar_data.matplotlib',fig_polar,dpi = 1600)
#fig_polar2 = heatPolar2D.get_fig(polar_data,grid = np.swapaxes(polar_grid,0,1))
#db.save(base_path + 'polar_data2.matplotlib',fig_polar2)

#ixs = np.arange(-5.5,6,0.5)
#intermediate_grid = GridFactory.construct_grid('uniform',[ixs,ixs])[:]

#tree = KDTree(pixel_centers.reshape(-1,2))
#tree2 = KDTree(pixel_centers.reshape(-1,2))
#q = tree.query(cart_grid.reshape(-1,2))
#q2 = tree.query(cart_grid.reshape(-1,2))
#ids = np.unravel_index(q[1],pixel_centers[:].shape[:-1])
#id2 = np.unravel_index(q[1],pixel_centers[:].shape[:-1])
#





#cart_pix_centers = spherical_to_cartesian(pixel_centers)[...,:2]
## create lookup for intermediate uniform grid    possible_point_ids[:,0]
#grid_intermediate_ids = (cart_grid+pixel_size/2)//pixel_size
#pix_intermediate_ids = (cart_pix_centers+pixel_size/2)//pixel_size
#
#intermediate_lookup_grid = generate_lookup_dict(grid_intermediate_ids)
#intermediate_lookup_pix = generate_lookup_dict(pix_intermediate_ids)
#
#
#
#lookup_array = create_lookup_array(pixel_centers,polar_grid,pixel_size)
#
#polar_data = data[lookup_array[...,0],lookup_array[...,1]]
#





import numpy as np
from xframe.settings import general
general.n_control_worker=0
import xframe
xframe.setup_experiment('fxs3046_online','default_proc')
exp = xframe.experiment_worker
exp.set_ids_for_run(169)

frames = np.arange(exp.nframes)
mask = (exp.train_ids>exp.train_ids.min() + 100) & (exp.train_ids < exp.train_ids.max() -100)
mask2 = exp.cell_ids != 0
frames = frames[mask & mask2]
a = xframe.lib.python.split_into_simple_slices(frames,mod = False,return_sliced_args = True)

#frames[-2] = 109324-202
#frames[-1] = 109324
#slices = []
#lengths = []
#jumpsl = []
#mod = 202
#eq_class_ids = frames%mod
#for c_id in range(mod):
#    eq_class = frames[eq_class_ids==c_id]
#    if len(eq_class)<=0:
#        continue
#    eq_min = eq_class[0]
#    temp = (eq_class-eq_min)//mod
#    jumps = np.nonzero(np.diff(temp)!=1)[0]+1
#    jumpsl.append(jumps)
#    connected_components = np.split(eq_class,jumps)
#    for c in connected_components:
#a
#if len(c) >0:
#            slices.append(slice(c[0],c[-1]+mod/2,mod))
#            lengths.append(len(c))
#



import logging
import sys
import os
import numpy as np
import traceback
from scipy.stats import unitary_group
file_path = os.path.realpath(__file__)
plugin_dir = os.path.dirname(file_path)
os.chdir(plugin_dir)


#from analysisLibrary.classes import ReciprocalProjectionData
from .projectLibrary import fxs_invariant_tools as i_tools
from .projectLibrary.ft_grid_pairs import max_order_from_n_angular_steps, get_grid
from .projectLibrary.fourier_transforms import generate_ft,load_fourier_transform_weights
from .projectLibrary.harmonic_transforms import HarmonicTransform
from .projectLibrary.hankel_transforms import generate_weightDict
from .projectLibrary.misk import generate_calc_center
from .projectLibrary.fxs_Projections import generate_shift_by_operator
from xframe.library.gridLibrary import SampledFunction,NestedArray,GridFactory
from xframe.library.mathLibrary import nearest_positive_semidefinite_matrix
from xframe.library.pythonLibrary import xprint
from xframe.library.mathLibrary import polar_spherical_dft_reciprocity_relation_radial_cutoffs,distance_from_line_2d
from xframe.interfaces import ProjectWorkerInterface
from xframe import database,settings
from xframe.library.mathLibrary import SampleShapeFunctions
from .projectLibrary.misk import _get_reciprocity_coefficient
from xframe import Multiprocessing
from xframe.library.mathLibrary import spherical_to_cartesian
log=logging.getLogger('root')

#class Worker(RecipeInterface):
class ProjectWorker(ProjectWorkerInterface):
    def __init__(self):
        self.extractor = InvariantExtractor()
    #       global db
    #        db = database.project
    def run(self):
        db = database.project
        opt = settings.project
        
        self.extractor.extract()

        
        if self.extractor.success:
            xprint('Saving Results \n')
            db.save('invariants', self.extractor)
        return {},locals()


######### InvariantExtractor #########
class InvariantExtractor:
    def __init__(self):
        self.dimensions = False
        self.xray_wavelength = False
        self.enforce_cross_correlation_qq_sym = False
        self.enforce_cross_correlation_phi_sym = False    
        self.average_intensity = False
        self.integrated_intensity = False
        self.data_projection_matrices = {'I1I1':False}
        self.data_low_resolution_intensity_coefficients = False
        self.pi_in_q = None
        self.reciprocity_coefficient = np.pi
        self.max_order = False
        self.data_n_particles = False
        self.data_radial_points = False
        self.data_angular_points = False
        self.data_max_q = False
        self.data_min_q = False
        self.proj_min_q = {'I1I1':False}
        self.number_of_particles = False
        self.extraction_routines = {
            'cross_correlation': self.extract_bl_from_cc,
            #'pdb': self.pre_init_from_pdb,
            'shapes': self.extract_bl_from_shapes,
            'density': self._bl_from_density
        }
        self.b_coeff={'I1I1':False}
        self.b_coeff_masks={'I1I1':False}
        self.b_coeff_q_id_limits = {'I1I1':False}
        self.data_projection_matrices_masks = {'I1I1':False}
        self.data_projection_matrices_q_id_limits = {'I1I1':False}
        self.data_projection_matrix_error_estimates={'I1I1':False}
        self.success=False

            
    def set_standard_instance_variables(self):
        opt = settings.project
        self.dimensions = opt.dimensions
        self.max_order = opt.max_order
        

    #################################
    ###    extraction routines    ###
    def extract_bl_from_cc(self):
        log.info('Computing degree 2 invariants from cross-correlation data:')
        db = database.project
        ccd = db.load('ccd')
        log.info('loaded ccd')                
        log.info(f'ccd shape = {ccd["cross_correlation"]["I1I1"].shape}')
        opt = settings.project
        average_intensity = ccd.get('average_intensity',False)

        self.average_intensity = average_intensity
        self.data_average_intensity = np.array(average_intensity.data)
        self.xray_wavelength = ccd['xray_wavelength']
        self.data_radial_points = ccd['radial_points']
        self.data_max_q = np.max(self.data_radial_points)
        self.data_min_q = np.min(self.data_radial_points)
        self.data_angular_points = ccd['angular_points']
        
        max_theoretical_order = len(self.data_angular_points)//2 #max_order_from_n_angular_steps(self.dimensions,len(self.data_angular_points),anti_aliazing_degree = 1)
        max_order = opt.max_order
        try:
            assert max_theoretical_order >= max_order , 'Violation of:  max specified harmonic order ({})<{} for the given number of angular points {}. Continue with max harmonic order = '.format(max_order,max_theoretical_order,len(self.data_angular_points),max_theoretical_order)
        except AssertionError as e:
            log.warning(e)
            max_order = max_theoretical_order
        self.max_order = max_order

        cc_arrays = ccd.pop('cross_correlation')
        for dset_name,dopt in opt.cross_correlation.datasets.items():
            if isinstance(opt.cross_correlation.datasets_to_process,(list,tuple)):
                if not (dset_name in opt.cross_correlation.datasets_to_process):
                    continue
            if dset_name in cc_arrays:
                if opt.dimensions==3:
                    extraction_method = dopt.bl_extraction_method
                elif opt.dimensions == 2:
                    extraction_method = '2d Fourier series'
                xprint(f'\tProcessing cross-correlation {dset_name} (extraction method = {extraction_method})...')
                
                cc = cc_arrays.pop(dset_name)
                metadata = {**ccd,**dopt,'orders':np.arange(self.max_order+1),'mode':dopt.bl_extraction_method,'average_intensity':self.data_average_intensity}
                #log.info('max eig 3 = {}'.format(np.linalg.eigh(np.fft.rfft(cc,axis=-1)[...,0])[0].max()))
                b_coeff,q_mask = i_tools.cross_correlation_to_deg2_invariant(cc,self.dimensions,**metadata)
                del(cc)
                #log.info(f"b coeff shape  =  {b_coeff.shape}")
                #log.info(dopt)
                mask,q_id_limits = self.calc_deg_2_invariant_masks(dopt,b_coeff.shape,q_mask)
                self.b_coeff_masks[dset_name] = mask
                self.b_coeff_q_id_limits[dset_name] = q_id_limits
                self.b_coeff[dset_name] = self.apply_invariant_constraints(dopt,b_coeff,q_id_limits)
                
        if ('I2I1' in self.b_coeff) and ('I1I1' in self.b_coeff) and ('I2I2' in self.b_coeff) :
            n_q1 = self.b_coeff_masks["I1I1"].shape[0]
            min_q2 = self.b_coeff_masks["I1I1"].any(axis = -1)
            combined_mask = self.b_coeff_masks["I1I1"].any(axis = 1)[:,None,:]*self.b_coeff_masks["I2I2"].any(axis = 2)[:,:,None]
            q_id_mins = np.concatenate((self.b_coeff_q_id_limits['I2I2'][:,0,None,:],self.b_coeff_q_id_limits['I1I1'][:,0,None,:]),axis = 1)
            self.b_coeff_masks['I2I1'] = combined_mask
            self.b_coeff_q_id_limits['I2I1'] = q_id_mins
            b3 = self.b_coeff['I2I1']
            for o,b in enumerate(b3):
                #enforce that b3 is of the form A*B^dagger where A and B are of shape (Nq,2*o+1)
                u,s,vh = np.linalg.svd(b)
                num_o = 2*o+1
                #eig,_ = np.linalg.eig(b)
                #log.info(np.sort(eig)[::-1])
                b[:] = (u[:,:num_o]*s[:num_o])@vh[:num_o,:]
        if opt.cross_correlation.datasets.I1I1.modify_cc.get('subtract_average_intensity',False):
            if self.dimensions==3:
                # since the spherical harmonic of degree 0 is  Y^0_0 = 1/(2*np.sqrt(np.pi))
                # Another way to look at it is that 4*pi is the total angular surface of a unit sphere 
                factor = 4*np.pi
            elif self.dimensions ==2:
                factor = 1 # 2*np.pi
            self.b_coeff['I1I1'][0]=average_intensity[:,None]*average_intensity[None,:]*factor
        #log.info(f'max bls = \n {[np.abs(b).max() for b in self.b_coeff["I1I1"]]}\n')
        
    def extract_bl_from_shapes(self):        
        opt = settings.project        
        shape_opt = opt.shapes

        if shape_opt.GPU.use:
            settings.general.n_control_workers = shape_opt.GPU.n_gpu_workers
            Multiprocessing.comm_module.restart_control_worker()
        self.reciprocity_coefficient = _get_reciprocity_coefficient(shape_opt.fourier_transform)        
        #self.pi_in_q = shape_opt.fourier_transform.pi_in_q
        
        # create_grid
        centers = np.asarray(shape_opt['shapes']['centers'])
        log.info('centers = {}'.format(centers))
        sizes = np.asarray(shape_opt['shapes']['sizes'])
        types = np.asarray(shape_opt['shapes']['types'])
        density_values = np.asarray(shape_opt['shapes']['densities'])
        random_orientation = np.asarray(shape_opt['shapes']['random_orientation'])
        max_particle_radius = shape_opt.get('shape_size','not given')
        if not isinstance(max_particle_radius,(float,int)):
            max_particle_radius = np.max(centers[:,0]+sizes)
        else:
            max_particle_radius/=2 #since mag radius is half the size of the shape.
            
        oversampling = shape_opt.grid.oversampling
        if isinstance(shape_opt.grid.max_q,bool):
            max_r = oversampling*max_particle_radius
            n_radial_points = shape_opt.grid.n_radial_points
            max_q = polar_spherical_dft_reciprocity_relation_radial_cutoffs(max_r,n_radial_points,reciprocity_coefficient = self.reciprocity_coefficient)
        else:
            max_q = shape_opt.grid.max_q#*(self.reciprocity_coefficient/np.pi)
            n_radial_points = shape_opt.n_radial_points
            max_r = polar_spherical_dft_reciprocity_relation_radial_cutoffs(max_q,n_radial_points,self.reciprocity_coefficient)
        self.data_max_q = max_q
        log.info('maxR = {} maxQ ={} reciprocity_coefficient = {}'.format(max_r,max_q,self.reciprocity_coefficient))

        ht_opt = {'dimensions':self.dimensions,'max_order':self.max_order,**shape_opt.grid}
        log.info('max order = {}'.format(self.max_order))
        cht = HarmonicTransform('complex',ht_opt)
        #weight_dict = generate_weightDict_zernike_spherical(max_order,n_radial_points)        
        grid_opt={'dimensions':opt.dimensions,'type':shape_opt.fourier_transform.type,'max_q':max_q,'n_radial_points':n_radial_points,**cht.grid_param,'reciprocity_coefficient':self.reciprocity_coefficient}
        grid_pair = get_grid(grid_opt)
        if opt.dimensions==3:
            self.data_radial_points =  grid_pair.reciprocalGrid[:,0,0,0]
            self.data_angular_points = grid_pair.reciprocalGrid[0,0,:,2]
        elif opt.dimensions==2:
            self.data_radial_points =  grid_pair.reciprocalGrid[:,0,0]
            self.data_angular_points = grid_pair.reciprocalGrid[0,:,1]
        self.data_max_q = np.max(self.data_radial_points)

        log.info('grid shape = {}'.format(grid_pair.realGrid[:].shape))
        log.info(f'grid extends: min {spherical_to_cartesian(grid_pair.realGrid[:]).min()} max {spherical_to_cartesian(grid_pair.realGrid[:]).max()}')

        density = np.zeros(grid_pair.realGrid[:].shape[:-1],dtype = float)
        for shape_type,center,size,dval,rand_rot in zip(types,centers,sizes,density_values,random_orientation):
            log.info(f'\n t {type}\n c {center}\n s {size}\n d {dval}\n r {random_orientation}')
            if shape_type == 'sphere':
                norm = 'standard'
                f = SampleShapeFunctions.get_disk_function(size,lambda points: np.full(points.shape[:-1],dval),center=center,norm=norm,random_orientation=rand_rot,coordSys='spherical')
            elif shape_type == 'tetrahedron':
                f = SampleShapeFunctions.get_tetrahedral_function(size,lambda points: np.full(points.shape[:-1],dval),center=center)
            elif shape_type == 'cube':
                norm = 'inf'
                log.info('size = {}'.format(sizes[i]))
                f = SampleShapeFunctions.get_disk_function(size,lambda points: np.full(points.shape[:-1],dval),center=center,norm=norm,random_orientation=rand_rot)
            else:
                norm = 'inf'
                f = SampleShapeFunctions.get_disk_function(size,lambda points: np.full(points.shape[:-1],dval),center=center,norm=norm,random_orientation=rand_rot)
            density += f(grid_pair.realGrid[:])
        log.info(f'radial density = {density.mean(axis=(1,2))} shape = {density.shape}')
            
        log.info('densty shape = {} grid shape = {} , max_R = {}'.format(density.shape,grid_pair.realGrid[:].shape,grid_pair.realGrid[...,0].max()))

        self._bl_from_density(density,shape_opt,opt.structure_name,n_radial_points,grid_pair,cht)


    def generate_fourier_transforms(self, grid_pair, harm_trf):
        opt = settings.project
        dimensions = opt.dimensions
        if dimensions==3:
            r_max = grid_pair.realGrid[:,0,0,0].max()
        elif dimensions ==2:
            r_max = grid_pair.realGrid[:,0,0].max()
        n_radial_points = grid_pair.reciprocalGrid.shape[0]
        
        weights_dict = load_fourier_transform_weights(opt.dimensions,opt.shapes.fourier_transform,opt.shapes.grid,database.project)

        ft_opt = opt.shapes.fourier_transform
        ft_type = ft_opt['type']
        #pi_in_q = ft_opt['pi_in_q']
        reciprocity_coefficient = _get_reciprocity_coefficient(ft_opt)
        use_gpu = opt.shapes.GPU.use
        log.info('reciprocity_coefficient = {}'.format(reciprocity_coefficient))
        fourierTransform,inverseFourierTransform=generate_ft(r_max,weights_dict,harm_trf,dimensions,reciprocity_coefficient=reciprocity_coefficient,use_gpu = use_gpu,mode = ft_type)      
        return fourierTransform,inverseFourierTransform

    def _bl_from_density(self,density,opt,name,n_radial_points,grid_pair,cht):
        self.reciprocity_coefficient = _get_reciprocity_coefficient(opt.fourier_transform)        
        db = database.project
        ft,ift = self.generate_fourier_transforms(grid_pair,cht)
        #log.info('start b_coeff calculation')
        #log.info('density shape = {}'.format(density.shape))
        #log.info('grid shape = {}'.format(grid_pair.real.shape))
        
        if settings.project.dimensions ==2:
            grid_type = 'polar'
        elif settings.project.dimensions ==3:
            grid_type = 'spherical'
        
        if opt.save_vtk_density:
            log.info(f'saving density {type(density)}')
            db.save('model_density', [density], grid = grid_pair.realGrid[:], grid_type=grid_type,path_modifiers={'name':name})
            log.info(f'alive')
        if opt.save_vtk_intensity:
            db.save('model_intensity', [np.abs(ft(density.astype(complex))).real], grid = grid_pair.reciprocalGrid[:], grid_type=grid_type,path_modifiers={'name':name})
        
        log.info('max_density = {}'.format(density.max()))
        n_particles=opt.n_particles
        
        self.b_coeff['I1I1'] = i_tools.density_to_deg2_invariants(density.astype(complex),ft,settings.project.dimensions,cht=cht)
        bl_shape = self.b_coeff['I1I1'].shape
        self.b_coeff_masks['I1I1'] = np.full(bl_shape,True)
        q_id_limits = np.zeros((bl_shape[0],)+(2,2),dtype=int)
        q_id_limits[...,1]=bl_shape[0]
        self.b_coeff_q_id_limits={'I1I1': q_id_limits}
        #self.b_coeff['I2I2'] = i_tools.density_to_deg2_invariants((density**2).astype(complex),ft,settings.project.dimensions,cht=cht)
        #self.b_coeff_masks['I2I2'] = np.full(self.b_coeff['I2I2'].shape,True)
        
        #self.b_coeff['I2I1'] = i_tools.density_to_deg2_invariants((density**2).astype(complex),ft,settings.project.dimensions,density2 =density.astype('complex'),cht=cht)
        #self.b_coeff_masks['I2I1'] = np.full(self.b_coeff['I2I1'].shape,True)
        #self.b_coeff[0] *= n_particles
        #log.info(self.b_coeff.shape)

        if settings.project.dimensions ==2:
            self.average_intensity = SampledFunction(NestedArray(self.data_radial_points,1),np.sqrt(np.diag(self.b_coeff[0])),coord_sys='cartesian')
        elif settings.project.dimensions ==3:
            average_data = np.sqrt(np.diag(self.b_coeff['I1I1'][0]).real/(4*np.pi))
            self.average_intensity = SampledFunction(NestedArray(self.data_radial_points,1),average_data,coord_sys='cartesian')

    def from_data(self,data):
        self.dimensions = data['dimensions']
        self.xray_wavelength = data['xray_wavelength']
        self.average_intensity = data['average_intensity']
        #log.info('aint type ={}'.format(type(self.average_intensity)))
        self.data_radial_points = data['data_radial_points'][:]
        self.data_angular_points = data['data_angular_points'][:]
        self.data_max_q = np.max(self.data_radial_points)
        self.data_min_q = data.get('data_min_q',False)
        #self.pi_in_q = data.get('pi_in_q',False)
        self.pi_in_q = data.get('pi_in_q',True)
        self.max_order = data['max_order']
        self.number_of_particles=data.get('number_of_particles',1)
        self.data_projection_matrices = data['data_projection_matrices']
        log.info('data keys = {}'.format(data.keys()))
        self.b_coeff = data['b_coeff']
        if isinstance(self.b_coeff,bool):
            self.b_coeff = i_tools.projection_matrices_to_deg2_invariant_3d(self.data_projection_matrices)
        self.success =True

        
    ###########################################################
    ###    PSD constraint + Projection matrix extraction    ###

    def calc_deg_2_invariant_masks(self,dopt,bl_shape,q_mask):
        min_type =  dopt.bl_q_limits.min.type
        max_type =  dopt.bl_q_limits.max.type

        empty_mask = np.ones(bl_shape,dtype = bool)
        #log.info(dopt.bl_q_limits)
            
        # min_limit        
        q_id_limits = np.zeros((bl_shape[0],)+(2,2),dtype=int)
        q_id_limits[...,1]=len(q_mask)
        if min_type == 'line':
            min_mask,q_id_mins = self.calc_deg_2_invariant_line_mask(dopt.bl_q_limits.min[min_type])
            q_id_limits[:,:,0]=q_id_mins
        else:
            min_mask = empty_mask.copy()
            
        # max_limit        
        if max_type == 'line':
            max_mask,q_id_maxs = self.calc_deg_2_invariant_line_mask(dopt.bl_q_limits.max[max_type],invert = True)
            q_id_limits[:,:,1]=q_id_maxs
        else:
            max_mask = empty_mask.copy()


        q_id_min = np.argmax(q_mask)
        q_id_max = np.argmax(q_mask[::-1])
        q_id_max = len(q_mask)-q_id_max
        mask = min_mask & max_mask
        mask[:,~q_mask]=False
        q_id_limits[q_id_limits[:,:,0]<q_id_min]=q_id_min
        q_id_limits[q_id_limits[:,:,1]>q_id_max]=q_id_max
        #log.info(f'computed q_ids shape = {q_id_limits.shape}')
        return mask,q_id_limits
        
        
        
    def calc_deg_2_invariant_line_mask(self,line_specifier,invert=False):
        qs = self.data_radial_points
        n_qs = len(qs)
        orders = np.arange(self.max_order+1)
        data_grid = GridFactory.construct_grid('uniform',[orders,qs])
        #log.info(f'line specifyer type = {type(line_specifier)}')
        if isinstance(line_specifier,tuple):
            mask1 = (-1*distance_from_line_2d(np.array(line_specifier[0]),data_grid[:]))>=0
            #log.info(f'mask1 shape = {mask1.shape}')
            mask2 = (-1*distance_from_line_2d(np.array(line_specifier[1]),data_grid[:]))>=0
            if not invert:
                q1_id = np.argmax(mask1,axis = 1)
                q2_id = np.argmax(mask2,axis = 1)
                if not mask1.any():
                    q1_id = n_qs-1
                if not mask2.any():
                    q2_id = n_qs-1
            else:
                mask1 = ~mask1
                mask2 = ~mask2
                q1_id = np.argmin(mask1,axis = 1)                
                q2_id = np.argmin(mask2,axis = 1)
                if mask1.all():
                    q1_id = n_qs
                if mask1.all():
                    q2_id = n_qs

            #log.info(f'mask2 shape = {mask2.shape}')
            q_id_limits = np.stack((q1_id,q2_id),axis = -1) 
            mask = mask1[:,:,None] * mask2[:,None,:]
        else:
            mask = (-1*distance_from_line_2d(np.array(line_specifier),data_grid[:]))>=0
            #log.info(f'mask shape = {mask.shape}')
            if not invert:
                q_id = np.argmax(mask,axis = 1)
                if not mask.any():
                    q_id = n_qs -1
            else:
                mask = ~mask
                q_id = np.argmin(mask,axis = 1)
                if mask.all():
                    q_id = n_qs
                #q_id = n_qs -q_id
            mask = mask[:,:,None] * mask[:,None,:]
            q_id_limits = np.stack((q_id,q_id),axis = -1) 
        #log.info(f'mask any = {mask.any()} mask all = {mask.all()}')
        return mask,q_id_limits


    def apply_invariant_constraints(self,dopt,b_coeff,q_id_limits):
        b_coeff_out = b_coeff.copy()
        if dopt.bl_enforce_psd:
            try:
                assert (q_id_limits[:,0,:]==q_id_limits[:,1,:]).all(),'Trying to compute projection matrices from non-square submatrix of deg2 invariants specified. That makes no sense since projection matrices are theoretically given by eig values of a positive semidefinite matrix. Continue using the q1_id_limits also for q2 to make the selection square.'
            except AssertionError as e:
                log.info(e)
                q_id_limits[:,1]=q_id_limits[:,0]
            for o in range(len(b_coeff)):
                q_slice = slice(*q_id_limits[o,0])
                b_coeff_out[o,q_slice,q_slice] = nearest_positive_semidefinite_matrix(b_coeff[o,q_slice,q_slice],low_positive_eigenvalues_to_zero=False)
            #eig,_=np.linalg.eigh(b_coeff_out[2])
            #log.info(f'min eigenvalue after corrections = {np.min(eig)}')
        return b_coeff_out
    

    def calc_projection_matrices(self):
        log.info('Extracting projection matrices:')
        log.info('Processing I1I1 ...')
        if settings.project.bl_eig_sort_mode == 'median_of_scaled_eigenvector':            
            sort_mode = 1
        else:
            sort_mode = 0
        
        temp = i_tools.deg2_invariant_to_projection_matrices(self.dimensions,self.b_coeff['I1I1'],q_id_limits=self.b_coeff_q_id_limits['I1I1'],sort_mode = sort_mode)
        self.data_projection_matrices['I1I1'],eig_I1I1 = temp
        self.data_projection_matrices_masks['I1I1'] = self.b_coeff_masks['I1I1'].sum(axis = 2)
        self.data_projection_matrix_error_estimates['I1I1'] = i_tools.calc_projection_matrix_error_estimate(self.b_coeff["I1I1"],self.data_projection_matrices['I1I1'])
        
        if settings.project.optimize_projection_matrices.use:
            log.info('Enforcing SHT constraints on projection matrices:')
            niter = settings.project.optimize_projection_matrices.n_iterations
            err_change_limit = settings.project.optimize_projection_matrices.error_change_limit
            new_proj_matrices = self.prephase_projection_matrices(niter,err_change_limit)
            self.data_projection_matrices['I1I1'] = new_proj_matrices
            
        if 'I2I2' in self.b_coeff:
            log.info('Processing I2I2 ...')
            temp = i_tools.deg2_invariant_to_projection_matrices(self.dimensions,self.b_coeff['I2I2'],q_id_limits = self.b_coeff_q_id_limits['I2I2'],sort_mode = sort_mode)
            self.data_projection_matrices['I2I2'],eig_I2I2 = temp
            self.data_projection_matrices_masks['I2I2'] = self.b_coeff_masks['I2I2'].sum(axis = 2)
            self.data_projection_matrix_error_estimates['I2I2'] = i_tools.calc_projection_matrix_error_estimate(self.b_coeff["I2I2"],self.data_projection_matrices['I2I2'])
            if 'I2I1' in self.b_coeff:
                log.info('Processing I2I1 ...')
                proj_I1I1 = self.data_projection_matrices['I1I1']
                proj_I2I2 = self.data_projection_matrices['I2I2']
                enforce_unitarity = settings.project.unitary_transform.enforce_unitarity
                if self.dimensions == 3:
                    self.data_projection_matrices['I2I1'],errors = i_tools.calc_unknown_unitary_transform(proj_I1I1,eig_I1I1,proj_I2I2,eig_I2I2,self.b_coeff['I2I1'],self.data_radial_points,q_id_limits = self.b_coeff_q_id_limits["I2I1"],method = settings.project.I2I1_unknown_tranrform_extraction_method)
                    self.data_projection_matrix_error_estimates['I2I1']=errors
    def calc_low_resolution_intensity_coefficients(self):
        max_order = settings.project.low_resolution_intensity_approximation.max_order
        if settings.project.optimize_projection_matrices.use:
            niter = settings.project.low_resolution_intensity_approximation.n_iterations
            err_change_limit = settings.project.low_resolution_intensity_approximation.error_change_limit
            new_projection_matrices = self.data_projection_matrices['I1I1'][:max_order+1]
            cht = HarmonicTransform('complex',{'dimensions':3,'max_order':max_order,'n_phi':0,'n_theta':0})
            new_proj_matrices,error_target_reached = i_tools.enforce_spherical_harmonic_transform_constraint(new_projection_matrices,niter,cht,rel_err_limit = err_change_limit)            
        else:
            new_proj_matrices = self.data_projection_matrices['I1I1'][:max_order+1]
        return new_proj_matrices
    
    def prephase_projection_matrices(self,n_iterations,err_change_limit):
        new_projection_matrices = [self.data_projection_matrices['I1I1'][0].copy()]
        for i in range(2,self.max_order,2):
            log.info(f'Processing orders <= {i}')
            new_projection_matrices.append(self.data_projection_matrices['I1I1'][i-1].copy())#@unitary_group.rvs(2*(i-1)+1))
            new_projection_matrices.append(self.data_projection_matrices['I1I1'][i].copy())#@unitary_group.rvs(2*i+1))
            #max_order = min(self.max_order,settings.project.low_resolution_intensity_approximation.max_order)
            cht = HarmonicTransform('complex',{'dimensions':3,'max_order':i,'n_phi':0,'n_theta':0})
            new_projection_matrices,error_target_reached = i_tools.enforce_spherical_harmonic_transform_constraint(new_projection_matrices,n_iterations,cht,rel_err_limit = err_change_limit)
        new_projection_matrices += self.data_projection_matrices['I1I1'][self.max_order:]
        cht = HarmonicTransform('complex',{'dimensions':3,'max_order':self.max_order,'n_phi':0,'n_theta':0})
        new_projection_matrices,error_target_reached = i_tools.enforce_spherical_harmonic_transform_constraint(new_projection_matrices,n_iterations,cht,rel_err_limit = err_change_limit)
        #cht = HarmonicTransform('complex',{'dimensions':3,'max_order':self.max_order,'n_phi':0,'n_theta':0})
        #new_proj_matrices,error_target_reached = i_tools.enforce_spherical_harmonic_transform_constraint(self.data_projection_matrices['I1I1'][:self.max_order+1],n_iterations,cht,rel_err_limit = 1e-6)
        return new_projection_matrices
    ##################
    ## main routine ##
    def extract(self):
        opt = settings.project
        db = database.project
        self.set_standard_instance_variables()        
        mode = settings.project.extraction_mode
        
        try:
            #this extracts the invariants
            xprint('Extracting invariants')
            self.extraction_routines[mode].__func__(self)
            xprint('done.\n')
        except KeyError as e:
            traceback.print_exc()
            log.error(' Extraction mode {} not known. Known modes are "{}"'.format(mode,self.extraction_routines.keys()))
            raise e

        xprint('\nComputing projection matrices Vl|vn')
        self.calc_projection_matrices()
        xprint('done.\n')

        xprint('Computing total intensity.')
        if opt.dimensions==3:
            self.integrated_intensity = np.trapz(self.average_intensity.data * self.data_radial_points**2 , x = self.data_radial_points,axis = 0)*4*np.pi
        elif opt.dimensions==2:
            self.integrated_intensity = np.trapz(self.average_intensity.data * self.data_radial_points , x = self.data_radial_points,axis = 0)*4*np.pi
        xprint('done.\n')
        
        for key,value in self.b_coeff_q_id_limits.items():
            self.data_projection_matrices_q_id_limits[key]=value[:,0]

        if self.dimensions==3:            
            low_res_Ilm = self.calc_low_resolution_intensity_coefficients()    
            self.data_low_resolution_intensity_coefficients = low_res_Ilm
            
        self.success =True
        #log.info(self.data_projection_matrices['I1I1'])
        xprint('Extraction completed! \n')
        



import numpy as np
import os
from xframe.plugins.MTIP.analysisLibrary.fourier_transforms import generate_ft
from xframe.plugins.MTIP.analysisLibrary.harmonic_transforms import HarmonicTransform
from xframe.plugins.MTIP.analysisLibrary.hankel_transforms import generate_weightDict,calc_spherical_trapz_weights,generate_weightDict_trapz,generate_ht,assemble_weights_trapz
from xframe.plugins.MTIP.analysisLibrary.ft_grid_pairs import get_grid
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import density_to_deg2_invariants
#from xframe.library.mathLibrary import SampleShapeFunctions
#from xframe.startup_routines import load_recipes
#from xframe.control.Control import Controller
#analysis_sketch = load_recipes(ana_name='MTIP.reconstruct', ana_settings='3d_pentagon_1p_00')[0]
#controller = Controller(analysis_sketch)
#from xframe.settings import analysis as opt
#from xframe.database import analysis as db

Q=0.41
n_r=128
max_order=64

ht2d = HarmonicTransform('complex',{'dimensions':2,'max_order':max_order})
grid_param = ht2d.grid_param
g = get_grid({'dimensions':2,'type':'trapz','pi_in_q':True,'max_q':Q,'n_radial_points':n_r,'phis':grid_param['phis']})
R = g.realGrid[:,0,0].max()

w = generate_weightDict(max_order,n_r,dimensions=2,mode='trapz',n_cpus=False,pi_in_q=True)
aw = assemble_weights_trapz(w['weights'],np.arange(max_order+1),R,True)
hk,ihk = generate_ht(w['weights'],np.arange(max_order+1),R,dimensions=2,pi_in_q=True)
ft,ift = generate_ft(R,w,ht2d,2,pi_in_q=True)
d = np.ones_like(g.realGrid.array[...,0])
hd = ht2d.forward(d)
d2=ift(ft(d))

print(np.sum(aw['inverse']*np.sum(aw['forward'],axis=0)[1:,None,:],axis=0)[:,0])






import numpy as np
from multiprocessing import Pool, cpu_count
from itertools import repeat
import time, math
import struct
import sys
import os
import scipy as sp
import scipy.stats as spst
#from scipy.fft import irfft,rfft
from numpy.fft import irfft,rfft
import logging
log = logging.getLogger("root")
Pi = math.pi

# cross-correlation analysis of the images
class ccf_analysis:    
    
    def __init__(self, n_q1, n_q2, n_phi, q1vals_pos, q2vals_pos):
        self.n_q1=n_q1 
        self.n_q2=n_q2 
        self.n_phi=n_phi 
        self.q1vals_pos=q1vals_pos    
        self.q2vals_pos=q2vals_pos
        self.fc_ccf_q1q2 = np.zeros((self.n_q1, self.n_q2, self.n_phi//2+1), dtype=np.complex128)
                 
    # calculate CCF and its FCs for a set of q1 != q2 rings. output is a (n_q1 * n_q * n_phi) matrix of CCF and FCs
    #
    def ccf_twopoint_q1_q2(self, data_polar1):
        #log.info(f'data contains Nan= {np.isnan(data_polar1).any()}')
        fc_I1q = rfft(data_polar1)
        np.multiply(np.conjugate(fc_I1q[self.q1vals_pos, None, :]), fc_I1q[None,self.q2vals_pos, :], out=self.fc_ccf_q1q2)
        ccf_q1q2=irfft(self.fc_ccf_q1q2,self.n_phi) #3D array, first index - q1, second - q2, third - 'angle phi'
        #log.info(f'correlation contains Nans = {np.isnan(ccf_q1q2).any()}')
        return  ccf_q1q2 

    # correct the CCF of the data by the CCF of the mask  
    #      
    def ccf_mask_correction(self, ccf_data, ccf_mask):
        ccf_data=ccf_data.real
        ccf_mask=ccf_mask.real
        nonzero_mask = (ccf_mask!=0)
        np.divide(ccf_data, ccf_mask, out=ccf_data, where=nonzero_mask)
        return ccf_data,nonzero_mask

    
    # Fourier components of the CCF 
    #
    def ccf_fcs(self, ccf_data):
        ccf_fcs=np.fft.fft(ccf_data)
        return ccf_fcs
    
    
    # mask-corrected two-point ccf 
    # 
    def ccf_twopoint_q1_q2_mask_corrected(self, image_pol, mask_pol):
        #log.info(f'data contains Nan= {np.isnan(image_pol).any()}')
        #log.info(f'no values are masked= {mask_pol.any()}')
        ccf_data=self.ccf_twopoint_q1_q2(image_pol)
        ccf_mask=self.ccf_twopoint_q1_q2(mask_pol)
        ccfcorrected,correction_mask=self.ccf_mask_correction(ccf_data, ccf_mask)
        return ccfcorrected,correction_mask
               
                           
    # perform symmetrization of the two-point ccf; used to correct for noisy at delta=0 or delta=2pi in the case of flat Ewald sphere
    #                       
    def symmetrize_ccf(self, ccf, posPi2, posPi, pos3Pi2):
        ccf_symmetric=np.empty_like(ccf)
        ccf_symmetric[...]=ccf[...]
        Nq2=ccf.shape[0]
        Nq1=ccf.shape[1]
        Nphi=ccf.shape[2]
        for i1 in range(Nq2):
            for j1 in range(Nq1):   
                ccf_symmetric[i1,j1,0:posPi2]=ccf[i1,j1,posPi:posPi+posPi2]
                ccf_symmetric[i1,j1,pos3Pi2+1:Nphi]=ccf[i1,j1,pos3Pi2+1-posPi:Nphi-posPi]
    
        return ccf_symmetric                        
                                                              



import logging
import glob
import os
from xframe.startup_routines import DefaultProjectArgparser
from xframe.startup_routines import DefaultProjectClick
from xframe.library.pythonLibrary import xprint
import xframe

log=logging.getLogger('root')

class ProjectArgparser(DefaultProjectArgparser):
    def __init__(self):
        super().__init__()
        self.project_description = 'Toolkit for the analysis of fluctuation X-ray scattering (FXS) experiments.'
    def add_correlate_argparser(self,subparser,worker):
        subparser.add_parser(worker, description='computes angular cross-correlations',help=f'computes angular cross-correlations',parents=[self.settings_parser])
    def add_extract_argparser(self,subparser,worker):
        subparser.add_parser(worker, description='extracts rotational invarants from cross-correlation data',help=f'extracts rotational invarants from cross-correlation data',parents=[self.settings_parser])
    def add_reconstruct_argparser(self,subparser,worker):
        subparser.add_parser(worker, description='phase retrieval based on rotational invariants',help=f'phase retrieval based on rotational invariants',parents=[self.settings_parser])
    def add_average_argparser(self,subparser,worker):
        parser = subparser.add_parser(worker, description='aligns and averages phase retrieval results',help=f'aligns and averages phase retrieval results',parents=[self.settings_parser])
    def add_reconstruct2_argparser(self,subparser,worker):
        pass
    def add_simulate_ccd_argparser(self,subparser,worker):
        subparser.add_parser(worker, description='Only for testing: generates ccd for simple shapes',help=f'Only for testing: generates ccd for simple shapes',parents=[self.settings_parser])        



class ProjectClick(DefaultProjectClick):
    def __init__(self):
        super().__init__()
        self.project_description = 'Toolkit for the analysis of fluctuation X-ray scattering (FXS) experiments.'
        self.short_description = 'Fluctuation X-ray scattering toolkit'

        self.worker_short_help = {
            'correlate':'computes angular cross-correlations ',
            'extract':'extracts rotational invarants',
            'reconstruct':'runs phase retrieval',
            'average':'aligns and averages reconstructions',
            'simulate_ccd':"testing: simulates cc's for simple shapes"
        }
        
        self.worker_help = {
            'correlate':'Computes the averaged angular cross-correlation function C(q1,q2,\delta) for a set of diffraction patterns. A settings file name must be provided via SETTINGS_NAME.',
            'extract':'Extracts rotational invariants Bl(q1,q2) from an averaged angular cross-correlation function and computes the projection matrices Vl|vn necessary for phase retrieval. A settings file name must be provided via SETTINGS_NAME.',
            'reconstruct':'Reconstructs the single particle electron density using the MTIP scheme for iterative phase retrieval. A settings file name must be provided via SETTINGS_NAME.',
            'average':'Aligns and averages multiple reconstructions. A settings file name must be provided via SETTINGS_NAME.',
            'simulate_ccd': 'Testing/Tutorial only: This tool allows to calculate averaged cross-correlation functions C(q1,q2,\delta) from simple shapes (known electron densities). A settings file name must be provided via SETTINGS_NAME.'
        }
        self.worker_epilog = {
            'correlate':'',
            'extract':'',
            'reconstruct':'',
            'average':'',
            'simulate_ccd':''
        }

    def add_fxs(self,group,click):
        @group.group(self.project_name,chain=True,help = self.project_description, short_help= self.short_description,invoke_without_command=True)
        @click.option('--restore_defaults', is_flag = True,is_eager=True,help='Restores default settings for all commands.')
        @click.pass_context
        def project(ctx,**kwargs):
            if kwargs['restore_defaults']:
                db = xframe.database.default
                workers = xframe.startup_routines.lookup_workers(xframe.known_projects['fxs'])
                for worker in workers:
                    project_path_modifier = {'project':'fxs','worker':worker}
                    settings_install_path = db.get_path('settings_install_project',path_modifiers=project_path_modifier,is_file=False)
                    backup_default_settings_files = glob.glob(os.path.join(settings_install_path,'backup_default_*.yaml'))
                    default_settings_files = [os.path.join(os.path.dirname(p),os.path.basename(p)[7:]) for p in backup_default_settings_files]
                    for orig,target in zip(backup_default_settings_files,default_settings_files):
                        log.info(f'Copying {orig} to {target}')
                        db.copy(orig,target)
                xprint(f'Default settings restored for {workers}')                
            else:
                if ctx.invoked_subcommand is None:
                    print(ctx.get_help())
        #project = self.add_default_project_click(group,click)
        return project
    def add_worker_correlate(self,group,click,worker):
        short_help=self.worker_short_help['correlate']
        help=self.worker_help['correlate']
        self.add_default_worker_click(group,click,worker,short_help=short_help,help=help)
    def add_worker_extract(self,group,click,worker):
        short_help=self.worker_short_help['extract']
        help=self.worker_help['extract']
        self.add_default_worker_click(group,click,worker,short_help=short_help,help=help)
    def add_worker_reconstruct(self,group,click,worker):
        short_help=self.worker_short_help['reconstruct']
        help=self.worker_help['reconstruct']
        self.add_default_worker_click(group,click,worker,short_help=short_help,help=help)
    def add_worker_average(self,group,click,worker):
        short_help=self.worker_short_help['average']
        help=self.worker_help['average']
        self.add_default_worker_click(group,click,worker,short_help=short_help,help=help)
    def add_worker_reconstruct2(self,group,click,worker):
        pass
    def add_worker_simulate_ccd(self,group,click,worker):
        short_help=self.worker_short_help['simulate_ccd']
        help=self.worker_help['simulate_ccd']
        self.add_default_worker_click(group,click,worker,short_help=short_help,help=help)
 



from xframe.interfaces import ProjectWorkerInterface
from xframe.settings import project as opt
from xframe.database import project as db
import numpy as np

class ProjectWorker(ProjectWorkerInterface):
    def run(self):
        print(f'Hello {opt.name}, your random number is: {opt.random_number}')
        data = {'name':opt.name,'random_number':opt.random_number,'data':np.arange(10,dtype=float)}
        #path = '~/.xframe/data/hello_dat.h5'
        db.save('my_data',data)
        data2 = db.load('my_data')
        print(f'Loaded data = {data2}')



from xframe.library.interfaces import PeakDetectorInterface
import logging


log=logging.getLogger('root')

"""UnionFind

Union-find data structure. Based on Josiah Carlson's code,
http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/215912
with significant additional changes by D. Eppstein.
"""

class UnionFind:
    """Union-find data structure.

    Each unionFind instance X maintains a family of disjoint sets of
    hashable objects, supporting the following two methods:

    - X[item] returns a name for the set containing the given item.
      Each set is named by an arbitrarily-chosen one of its members; as
      long as the set remains unchanged it will keep the same name. If
      the item is not yet part of a set in X, a new singleton set is
      created for it.

    - X.union(item1, item2, ...) merges the sets containing each item
      into a single larger set.  If any item is not yet part of a set
      in X, it is added to X as one of the members of the merged set.
    """
    def __init__(self):
        """Create a new empty union-find structure."""
        self.weights = {}
        self.parents = {}
        

    def add(self, object, weight):
        if object not in self.parents:
            self.parents[object] = object
            self.weights[object] = weight

    def __contains__(self, object):
        return object in self.parents

    def __getitem__(self, object):
        """Find and return the name of the set containing the object."""

        # check for previously unknown object
        if object not in self.parents:
            assert(False)
            self.parents[object] = object
            self.weights[object] = 1
            return object

        # find path of objects leading to the root
        path = [object]
        root = self.parents[object]
        while root != path[-1]:
            path.append(root)
            root = self.parents[root]

        # compress the path and return
        for ancestor in path:
            self.parents[ancestor] = root
        return root

    def __iter__(self):
        """Iterate through all items ever found or unioned by this structure.
        """
        return iter(self.parents)

    def union(self, *objects):
        """Find the sets containing the objects and merge them all."""
        roots = [self[x] for x in objects]
        heaviest = max([(self.weights[r], r) for r in roots])[1]
        for r in roots:
            if r != heaviest:
                self.parents[r] = heaviest



"""A simple implementation of persistent homology on 2D images."""

__author__ = "Stefan Huber <shuber@sthu.org>"

def get(im, p):
    return im[p[0]][p[1]]


def iter_neighbors(p, w, h,periodic = False):
    y, x = p

    # 8-neighborship
    #neigh = [(y+j, x+i) for i in [-1, 0, 1] for j in [-1, 0, 1]]
    # 4-neighborship
    neigh = [(y-1, x), (y+1, x), (y, x-1), (y, x+1)]
    #log.info(periodic)
    
    if periodic:
        for j, i in neigh:
            if j == y and i == x:
                continue
            yield j%h, i%w
    else:
        for j, i in neigh:
            if j < 0 or j >= h:
                continue
            if i < 0 or i >= w:
                continue
            if j == y and i == x:
                continue
            yield j, i
            
    

def get_persistent_homology_2d(im,periodic=False):
    h, w = im.shape

    # Get indices orderd by value from high to low
    indices = [(i, j) for i in range(h) for j in range(w)]
    indices.sort(key=lambda p: get(im, p), reverse=True)
    #log.info(indices[:5])

    # Maintains the growing sets
    uf = UnionFind()

    groups0 = {}

    def get_comp_birth(p):
        return get(im, uf[p])

    # Process pixels from high to low
    for i, p in enumerate(indices):
        v = get(im, p)
        ni = [uf[q] for q in iter_neighbors(p, w, h,periodic = periodic) if q in uf]
        nc = sorted([(get_comp_birth(q), q) for q in set(ni)], reverse=True)

        if i == 0:
            groups0[p] = (v, v, None)
        
        uf.add(p, -i)
            
        if len(nc) > 0:
            oldp = nc[0][1]
            uf.union(oldp, p)

            # Merge all others with oldp
            for bl, q in nc[1:]:
                if uf[q] not in groups0:
                    #print(i, ": Merge", uf[q], "with", oldp, "via", p)
                    groups0[uf[q]] = (bl, bl-v, p)
                uf.union(oldp, q)

    groups0 = [(k, groups0[k][0], groups0[k][1], groups0[k][2]) for k in groups0]
    groups0.sort(key=lambda g: g[2], reverse=True)

    return groups0



class Peak:
    def __init__(self, startidx):
        self.born = self.left = self.right = startidx
        self.died = None

    def get_persistence(self, seq):
        return float("inf") if self.died is None else seq[self.born] - seq[self.died]

def get_persistent_homology_1d(seq):
    peaks = []
    # Maps indices to peaks
    idxtopeak = [None for s in seq]
    # Sequence indices sorted by values
    indices = range(len(seq))
    indices = sorted(indices, key = lambda i: seq[i], reverse=True)

    # Process each sample in descending order
    for idx in indices:
        lftdone = (idx > 0 and idxtopeak[idx-1] is not None)
        rgtdone = (idx < len(seq)-1 and idxtopeak[idx+1] is not None)
        il = idxtopeak[idx-1] if lftdone else None
        ir = idxtopeak[idx+1] if rgtdone else None

        # New peak born
        if not lftdone and not rgtdone:
            peaks.append(Peak(idx))
            idxtopeak[idx] = len(peaks)-1

        # Directly merge to next peak left
        if lftdone and not rgtdone:
            peaks[il].right += 1
            idxtopeak[idx] = il

        # Directly merge to next peak right
        if not lftdone and rgtdone:
            peaks[ir].left -= 1
            idxtopeak[idx] = ir

        # Merge left and right peaks
        if lftdone and rgtdone:
            # Left was born earlier: merge right to left
            if seq[peaks[il].born] > seq[peaks[ir].born]:
                peaks[ir].died = idx
                peaks[il].right = peaks[ir].right
                idxtopeak[peaks[il].right] = idxtopeak[idx] = il
            else:
                peaks[il].died = idx
                peaks[ir].left = peaks[il].left
                idxtopeak[peaks[ir].left] = idxtopeak[idx] = ir

    # This is optional convenience
    return sorted(peaks, key=lambda p: p.get_persistence(seq), reverse=True)


class PersistentHomologyPlugin(PeakDetectorInterface):
    @staticmethod
    def find_peaks(dim,data,periodic=False):
        if dim == 1:
            return get_persistent_homology_1d(data)
        elif dim == 2:
            return get_persistent_homology_2d(data,periodic=periodic)
            



import os
os.environ["MKL_NUM_THREADS"] = "1" 
os.environ["NUMEXPR_NUM_THREADS"] = "1" 
os.environ["OMP_NUM_THREADS"] = "1" 
os.environ['OPENBLAS_NUM_THREADS'] = "1"


import numpy as np
import scipy
import logging
import traceback
from xframe import log
os.chdir(os.path.expanduser('~/Programs/xframe'))
from xframe.startup_routines import load_recipes
from xframe.startup_routines import dependency_injection_no_soft
from xframe.library import mathLibrary
from xframe.externalLibraries.flt_plugin import LegendreTransform
mathLibrary.leg_trf = LegendreTransform
from xframe.externalLibraries.shtns_plugin import sh
mathLibrary.shtns = sh
from xframe.control.Control import Controller
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import cross_correlation_to_deg2_invariant
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_apply_precision_filter
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import ccd_to_deg2_invariant_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import bl_3d_pseudo_inverse_worker
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_cc_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import density_to_deg2_invariants_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_projection_matrices
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_projection_matrices_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_eigenvalues
from xframe.plugins.MTIP.analysisLibrary import fxs_invariant_tools as i_tools
from xframe.plugins.MTIP.analysisLibrary.hankel_transforms import generate_weightDict_zernike_spherical
from xframe.plugins.MTIP.analysisLibrary.harmonic_transforms import HarmonicTransform
from xframe.plugins.MTIP.analysisLibrary.fourier_transforms import generate_zernike_spherical_ft
from xframe.plugins.MTIP.analysisLibrary.ft_grid_pairs import get_grid
from xframe.plugins.MTIP.analysisLibrary.fxs_Projections import ReciprocalProjection_old,ReciprocalProjection
from xframe import settings
from xframe.library import mathLibrary as mLib
from xframe import Multiprocessing
from xframe.library import physicsLibrary as pLib
log=log.setup_custom_logger('root','INFO')

dependency_injection_no_soft()
analysis_sketch = load_recipes(ana_name='MTIP.reconstruct', ana_settings='3d_model_0')[0]
db = analysis_sketch[1]
opt = settings.analysis
controller = Controller(analysis_sketch)


def test_cc_bl_pseudo_inverse():
    ccd= db.load('ccd')
    cc_data = ccd['cross_correlation']
    
    orders = opt.fourier_transform.pos_orders[opt.projections.reciprocal.used_order_ids]
    
    bl = cross_correlation_to_deg2_invariant(ccd['cross_correlation'],3,{**ccd,'orders':orders,'mode':'pseudo_inverse','cc_mask_threshold':-1})
    cc_inverted = deg2_invariant_to_cc_3d(bl,ccd['xray_wavelength'],ccd['data_grid'],orders)
    angles = ccd["phis"]<np.pi
    
    rel_diff = np.abs(cc_data-cc_inverted)[...,angles]/np.abs(cc_data[...,angles])
    print('Pseudo inverse ruslan_data: mean = {} std = {}'.format(np.mean(rel_diff),np.std(rel_diff)))
    return locals()


def test_cc_bl_legendre():
    ccd= db.load('ccd')
    cc_data = ccd['cross_correlation']
    
    orders = np.arange(0,100,2)
    
    bl2 = cross_correlation_to_deg2_invariant(ccd['cross_correlation'],3,{**ccd,'orders':orders,'mode':'legendre_approx','cc_mask_threshold':-1})
    cc_inverted2 = deg2_invariant_to_cc_3d(bl2,ccd['xray_wavelength'],ccd['data_grid'],orders)

    angles = ccd["phis"]<np.pi
    
    rel_diff = np.abs(cc_data-cc_inverted2)[...,angles]/np.abs(cc_data[...,angles])
    print('Legendre_approx ruslan_data: mean = {} std = {}'.format(np.mean(rel_diff),np.std(rel_diff)))
    return locals()

#locals().update(test_cc_bl_legendre())
#locals().update(test_cc_bl_pseudo_inverse())

def test_bl_pseudo_inverse():
    ccd= db.load('ccd')
    cc_data = ccd['cross_correlation']
    
    orders = opt.fourier_transform.pos_orders[opt.projections.reciprocal.used_order_ids]
    
    bl = cross_correlation_to_deg2_invariant(ccd['cross_correlation'],3,{**ccd,'orders':orders,'cc_mask_threshold':-1})
    cc_inverted = deg2_invariant_to_cc_3d(bl,ccd['xray_wavelength'],ccd['data_grid'],orders)
    bl2 = cross_correlation_to_deg2_invariant(cc_inverted,3,{**ccd,'orders':orders,'cc_mask_threshold':-1})
    angles = ccd["phis"]<np.pi

    nonzero = np.abs(bl)!=0
    rel_diff = np.abs(bl-bl2)[nonzero]/np.abs(bl[nonzero])
    print('Pseudo inverse ruslan_data: mean = {} std = {}'.format(np.mean(rel_diff),np.std(rel_diff)))
    return locals()
#locals().update(test_bl_pseudo_inverse())

#rel_diff_leg = np.abs((cc_data[...,:800] - cc_inverted2)/cc_data[...,:800])
#rel_diff_pseudo = np.abs((cc_data[...,:800] - cc_inverted)/cc_data[...,:800])

def test_bl_legendre():
    ccd= db.load('ccd')
    cc_data = ccd['cross_correlation']
    angles = ccd["phis"]<np.pi
    
    orders = opt.fourier_transform.pos_orders[opt.projections.reciprocal.used_order_ids][::2]
    
    bl = cross_correlation_to_deg2_invariant(ccd['cross_correlation'],3,{**ccd,'orders':orders,'mode':'legendre_approx','cc_mask_threshold':-1})
    cc_inverted = deg2_invariant_to_cc_3d(bl,ccd['xray_wavelength'],ccd['data_grid'],orders)
    rel_diff = np.abs(cc_data-cc_inverted)[...,angles]/np.abs(cc_data[...,angles])
    print('Legendre_approx ruslan_data: mean = {} std = {}'.format(np.mean(rel_diff),np.std(rel_diff)))
    bl2 = cross_correlation_to_deg2_invariant(cc_inverted,3,{**ccd,'orders':orders,'mode':'legendre_approx','cc_mask_threshold':-1})

    nonzero = np.abs(bl)!=0
    rel_diff = np.abs(bl-bl2)[nonzero]/np.abs(bl[nonzero])
    print('Pseudo inverse ruslan_data: mean = {} std = {}'.format(np.mean(rel_diff),np.std(rel_diff)))
    return locals()
#locals().update(test_bl_legendre())

def init_transforms():
    ccd= db.load('ccd')
    cc_data = ccd['cross_correlation'][...,:800]
    maxQ = np.max(ccd['radial_points'])
    max_order = 99
    n_radial_points = 128
    opt=settings.analysis
    ht_opt={'dimensions':3,'max_order':max_order,'indices':'lm','anti_aliazing_degree':2,'n_phi':0,'n_theta':0}
    #ht_opt={'dimensions':opt.dimensions,**opt['grid'],'type':opt['fourier_transform']['type'],'pi_in_q':opt['fourier_transform']['pi_in_q'],'reciprocal_cut_off':maxQ}
    # harmonic transforms and grid
    cht=HarmonicTransform('complex',ht_opt)
    weight_dict = generate_weightDict_zernike_spherical(max_order,n_radial_points)
    grid_opt={'dimensions':3,'type':'Zernike_dict','max_q':maxQ,'n_radial_points':n_radial_points,**cht.grid_param,'pi_in_q':True}
    grid_pair = get_grid(grid_opt)
    maxR = grid_pair.real[:,0,0].max()
    ft,ift = generate_zernike_spherical_ft(maxR,weight_dict,cht,'complex',True,use_gpu=True)
    return locals()


def generate_density_spheres(grid,radius = 100, norm = 'standard',cart_centers= np.array([[0,200,0],[0,-200,0],[200,0,0],[-200,0,0],[400,0,0],[0,0,300]],dtype = np.float64),random_orientation = False):
    centers = mLib.cartesian_to_spherical(cart_centers)
    #centers = [(100,0,np.pi),(100,np.pi,np.pi),(100,0,0),(200,np.pi,0),(300,np.pi,np.pi/2)]
    sphere_funcs = [mLib.SampleShapeFunctions.get_disk_function(radius,center = center,norm=norm,random_orientation = random_orientation) for center in centers]
    density = sphere_funcs.pop(0)(grid)
    for sphere_func in sphere_funcs:
        density += sphere_func(grid)
    return density

def test_cc_simulated(density, ht, ft, grid,max_order):
    orders = np.arange(max_order+1)
    radial_points = grid.reciprocal[:,0,0,0]
    angular_points = np.arange(1000)/1000*2*np.pi
    cc_data_grid = {'qs':radial_points,"phis":angular_points}    
    ccd= db.load('ccd')
    ccd['data_grid'] = cc_data_grid
    ccd['phis'] = cc_data_grid['phis']
    Bl = density_to_deg2_invariants_3d(density,ht,ft)
    print('bl shape = {}'.format(Bl.shape))
    cc = deg2_invariant_to_cc_3d(Bl,ccd['xray_wavelength'],cc_data_grid,orders)
    print('cc finished')
    Bl2 = cross_correlation_to_deg2_invariant(cc,3,{**ccd,'orders':orders,'mode':'legendre_approx','cc_mask_threshold':-1})
    print('bl2 finished')
    cc2 = deg2_invariant_to_cc_3d(Bl2,ccd['xray_wavelength'],ccd['data_grid'],orders)
    return locals()
#locals().update(init_transforms())
#density = generate_density_spheres(grid_pair.real,radius = 200,norm='inf',cart_centers = np.array([[-200,0,0],[200,0,0]],dtype=np.float64),random_orientation = True)
#locals().update(test_cc_simulated(density,cht,ft,grid_pair,max_order))
#angles = ccd["phis"]<np.pi
#rel_diff = np.abs((cc2-cc)/cc2)[...,angles]
#print('mean = {} std = {}'.format(np.mean(rel_diff),np.std(rel_diff)))
#locals().update(test_cc_bl_pseudo_inverse())

#a=test_cc_bl_pseudo_inverse()
#a=test_cc_bl_legendre()
    
def projection_matrix():
    try:
        ccd= db.load('ccd')
        cc = ccd['cross_correlation']
        sym_cc = i_tools.cross_correlation_enforce_symmetry(cc)
        ccd['cross_correlation'] = sym_cc
        orders = opt.fourier_transform.pos_orders[opt.projections.reciprocal.used_order_ids]
        metadata={**ccd,'orders':orders,'cc_mask_threshold':-1}
        cc_mask = i_tools.cross_correlation_mask(metadata['qs'],metadata['thetas'],metadata['phis'],threshold = metadata['cc_mask_threshold'])
        bl = cross_correlation_to_deg2_invariant(ccd['cross_correlation'],3,metadata)
        #bl[1::2,...]=0
        p = deg2_invariant_to_projection_matrices(3,bl)
        ReciprocalProjection_old.pre_init_from_ccd(ccd,Multiprocessing.comm_module,{**metadata,'max_order':orders.max(),'dimensions':3})
        p2 = ReciprocalProjection_old.data_projection_matrices

        if ~np.any([(new-old).any() for new,old in zip(p,p2)]):
            print('test passed!')
        
    except Exception as e:
        traceback.print_exc()
    return locals()

#locals().update(projection_matrix())


def projection_matrix_new():
    try:
        ccd= db.load('ccd')
        cc = ccd['cross_correlation']
        sym_cc = i_tools.cross_correlation_enforce_symmetry(cc)
        ccd['cross_correlation'] = sym_cc
        orders = opt.fourier_transform.pos_orders[opt.projections.reciprocal.used_order_ids]
        metadata={**ccd,'orders':orders,'cc_mask_threshold':-1}
        cc_mask = i_tools.cross_correlation_mask(metadata['qs'],metadata['thetas'],metadata['phis'],threshold = metadata['cc_mask_threshold'])
        bl = cross_correlation_to_deg2_invariant(ccd['cross_correlation'],3,metadata)
        #bl[1::2,...]=0
        p = deg2_invariant_to_projection_matrices(3,bl)
        ReciprocalProjection.pre_init_from_ccd(ccd)
        p2 = ReciprocalProjection.data_projection_matrices

        if ~np.any([(new-old).any() for new,old in zip(p,p2)]):
            print('test passed!')
        
    except Exception as e:
        traceback.print_exc()
    return locals()

#locals().update(projection_matrix_new())

def projection_reciprocal_projection():
    try:
        ccd= db.load('ccd')
        cc = ccd['cross_correlation']
        locals().update(init_transforms())
        sym_cc = i_tools.cross_correlation_enforce_symmetry(cc)
        ccd['cross_correlation'] = sym_cc
        ReciprocalProjection.pre_init_from_ccd(ccd)
        p2 = ReciprocalProjection.data_projection_matrices        
    except Exception as e:
        traceback.print_exc()
    return locals()

#locals().update(projection_reciprocal_projection())
#locals().update(init_transforms())
#pr = ReciprocalProjection(grid_pair.reciprocal)

def test_leg_pseudo():
    sys._getframe(1).f_locals.update(init_transforms())
    density = generate_density_spheres(grid_pair.real,radius = 200,norm='inf',cart_centers = np.array([[-200,0,0],[200,0,0]],dtype=np.float64),random_orientation = False)
    Bl = density_to_deg2_invariants_3d(density,cht,ft).real
    
    orders = np.arange(100)
    wavelength = pLib.energy_to_wavelength(10000)*1e10
    qs = grid_pair.reciprocal[:,0,0,0]
    phis = np.arange(1000)*2*np.pi/1000
    data_grid = {'qs':qs,'phis':phis,'thetas':pLib.ewald_sphere_theta_pi(wavelength,qs)}
    cc = deg2_invariant_to_cc_3d(Bl,wavelength,data_grid,orders)
    bl_pse = cross_correlation_to_deg2_invariant(cc,3,{**data_grid,'data_grid':data_grid,"xray_wavelength":wavelength,'orders':orders,'mode':'pseudo_inverse','cc_mask_threshold':-1})
    bl_leg = cross_correlation_to_deg2_invariant(cc,3,{**data_grid,"data_grid":data_grid,"xray_wavelength":wavelength,'orders':orders,'mode':'legendre_approx','cc_mask_threshold':-1})

    bl_nonzero = Bl!=0
    rel_pse = np.abs(Bl-bl_pse)/np.abs(Bl)
    rel_leg = np.abs(Bl-bl_leg)/np.abs(Bl)    
    return locals()


locals().update(test_leg_pseudo())

#a=test_cc_bl_pseudo_inverse()
#a=test_cc_bl_legendre()
    



import os
os.environ["MKL_NUM_THREADS"] = "1" 
os.environ["NUMEXPR_NUM_THREADS"] = "1" 
os.environ["OMP_NUM_THREADS"] = "1" 
os.environ['OPENBLAS_NUM_THREADS'] = "1"


import numpy as np
import scipy
import logging
import traceback
from xframe import log
os.chdir(os.path.expanduser('~/Programs/xframe'))
from xframe.startup_routines import load_recipes
from xframe.startup_routines import dependency_injection_no_soft
from xframe.library import mathLibrary
from xframe.externalLibraries.flt_plugin import LegendreTransform
mathLibrary.leg_trf = LegendreTransform
from xframe.externalLibraries.shtns_plugin import sh
mathLibrary.shtns = sh
from xframe.control.Control import Controller
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import cross_correlation_to_deg2_invariant
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import ccd_remove_0_order,cross_correlation_low_pass
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import cross_correlation_mask
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_apply_precision_filter
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import ccd_to_deg2_invariant_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import bl_3d_pseudo_inverse_worker
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_cc_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import density_to_deg2_invariants_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_projection_matrices
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_projection_matrices_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_eigenvalues
from xframe.plugins.MTIP.analysisLibrary import fxs_invariant_tools as i_tools
from xframe.plugins.MTIP.analysisLibrary.hankel_transforms import generate_weightDict_zernike_spherical
from xframe.plugins.MTIP.analysisLibrary.harmonic_transforms import HarmonicTransform
from xframe.plugins.MTIP.analysisLibrary.fourier_transforms import generate_zernike_spherical_ft
from xframe.plugins.MTIP.analysisLibrary.ft_grid_pairs import get_grid
from xframe.plugins.MTIP.analysisLibrary.fxs_Projections import ReciprocalProjection_old,ReciprocalProjection
from xframe import settings
from xframe.library import mathLibrary as mLib
from xframe import Multiprocessing
from xframe.library import physicsLibrary as pLib
from xframe.library.gridLibrary import GridFactory 
log=log.setup_custom_logger('root','INFO')

dependency_injection_no_soft()
#analysis_sketch = load_recipes(ana_name='MTIP.reconstruct', ana_settings='3d_model_0')[0]
analysis_sketch = load_recipes(ana_name='MTIP.reconstruct', ana_settings='3d_pentagon_100')[0] 
db = analysis_sketch[1]
opt = settings.analysis
controller = Controller(analysis_sketch)
from xframe.presenters.matplolibPresenter import heat2D,heat2D_multi

ccd = db.load('ccd')
ccd_1 = db.load_ccd('/gpfs/exfel/theory_group/user/berberic/MTIP/ccd/3d_model_0.h5')
cc = ccd['cross_correlation']
cc_1 = ccd_1['cross_correlation']

#new_cc,_locals =ccd_remove_0_order(cc)
#locals().update(_locals)
metadata = {}
proj_opt = settings.analysis.projections.reciprocal.pre_init.cross_correlation
extract_odd_orders = settings.analysis.projections.reciprocal.pre_init.extract_odd_orders
max_order = settings.analysis.projections.reciprocal.pre_init.max_order
#mask_opt = {'type': 'donatelli','donatelli':{'threshold': -1}}
#metadata = {**ccd,'orders':np.arange(max_order+1),'mask':proj_opt.mask,'n_particles':proj_opt.n_particles}        
#new_bl = cross_correlation_to_deg2_invariant(new_cc,3,metadata)[0]
metadata = {**ccd,'orders':np.arange(max_order+1),'mask':proj_opt.mask,'n_particles':1,'extract_odd_orders':False}    
bl = cross_correlation_to_deg2_invariant(cc_1,3,metadata)[0]

max_orders = [5,10,15,20,25,99]
#max_orders = [5,10]
n_orders = len(max_orders)
order = 2
grid = GridFactory.construct_grid('uniform',[ccd['qs'],ccd['qs']])
bl_2_l = [np.abs(bl[order]).real]
bl_2_0_f = [np.abs(bl[order]).real]
bl_2_t = [np.abs(bl[order]).real]
bl_2_f = [np.abs(bl[order]).real]
metadata = {**ccd,'orders':np.arange(max_order+1),'mask':proj_opt.mask,'n_particles':proj_opt.n_particles,'extract_odd_orders':False}    
cc_mask = cross_correlation_mask(ccd['data_grid'],metadata)
for max_order in max_orders:
    new_cc = np.copy(cc)
    new_cc[~cc_mask]=0
    new_cc,_locals = cross_correlation_low_pass(cc,max_order)   
    #metadata = {**ccd,'orders':np.arange(max_order+1),'mask':proj_opt.mask,'n_particles':proj_opt.n_particles,'extract_odd_orders':True}    
    #bl_2_t.append(np.abs(cross_correlation_to_deg2_invariant(cc,3,metadata)[0][order]).real)
    #bl_2_0_f.append(np.abs(cross_correlation_to_deg2_invariant(new_cc,3,metadata)[0][order]).real)
    bl_new = cross_correlation_to_deg2_invariant(cc_new,3,metadata)[0][order]
    bl_old = cross_correlation_to_deg2_invariant(cc,3,metadata)[0][order]
    bl_2_f.append(np.abs(bl_old).real)
    bl_2_l.append(np.abs(bl_new).real)
    


#ls = 0,2,4,6,8
#db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/cc_remove_zero_component/bl_without_0_100_mO11.matplotlib',heat2D_multi.get_fig([[np.abs(bl[l]).real for l in ls],[np.abs(new_bl[l]).real for l in ls]],scale = 'log',shape = [2,len(ls)],size = (30,10),vmin= 1e14, vmax = 1e24,cmap = 'plasma'))

layout_l = {'title':'$B_2$ low-pass-filtered','x_label':'$q_1$', 'y_label':'$q_2$','title_size':40,'text_size':30}
layout_f = {'title':'$B_2$','x_label':'$q_1$', 'y_label':'$q_2$','title_size':40,'text_size':30}
layout = tuple(zip(*[[layout_f]*(n_orders+1),[layout_l]*(n_orders+1)]))

db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/cc_remove_zero_component/b0_comparison_low_pass.matplotlib',heat2D_multi.get_fig(tuple(zip(*[bl_2_f,bl_2_l])),layout=layout, grid = grid,scale = 'log',shape = [len(max_orders)+1,2],size = (30,10*(len(bl_2_f)+1)),vmin= 1e14, vmax = 1e24,cmap = 'plasma'))



import logging
import numpy as np
from itertools import repeat
import ctypes as C
from ctypes import c_double,c_int,c_size_t,c_float

from xframe.externalLibraries.gsl_plugin import gslWrapper
from xframe.externalLibraries.gsl_plugin import gslStructures as gslStruct
from xframe.library.interfaces import GSLInterface

log=logging.getLogger('root')
gsl=gslWrapper.Wraper()
c_pint=C.POINTER(c_int)
c_pdouble=C.POINTER(c_double)
c_pfloat=C.POINTER(c_float)
c_psize_t=C.POINTER(c_size_t)

class FFTPrecomputedDataHandler:
    class FFTPrecomputedData:
        dataTypes={'complex':0,'real':1,'halfcomplex':2}
        
        def __init__(self,dataType):            
            dataTypes=self.dataTypes                
            try:
                if dataType==dataTypes['complex']:
                    generateDatasetFactoryArgs=[gsl.gsl_fft_complex_wavetable_alloc,gsl.gsl_fft_complex_workspace_alloc]
                    deleteDatasetFactoryArgs=[gsl.gsl_fft_complex_wavetable_free,gsl.gsl_fft_complex_workspace_free]
                elif dataType==dataTypes['real']:
                    generateDatasetFactoryArgs=[gsl.gsl_fft_real_wavetable_alloc,gsl.gsl_fft_real_workspace_alloc]
                    deleteDatasetFactoryArgs=[gsl.gsl_fft_real_wavetable_free,gsl.gsl_fft_real_workspace_free]
                elif dataType==dataTypes['halfcomplex']:
                    generateDatasetFactoryArgs=[gsl.gsl_fft_halfcomplex_wavetable_alloc]
                    deleteDatasetFactoryArgs=[gsl.gsl_fft_halfcomplex_wavetable_free]
                else:
                    raise AssertionError('FFTPrecomputedData does not know the data type = {} \n Known dataTypes are listed as class varibales starting with dataType_'.format(dataType))
            except AssertionError as error:
                log.error(error)

            self.dataType=dataType
            self.dataDict={'fftSizes':[],'precomputedData':[]}
            self.generateDataset=self.generateDataset_MethodFactory(*generateDatasetFactoryArgs)
            self.deleteSingleDataset=self.deleteDataset_MethodFactory(*deleteDatasetFactoryArgs)


        def getDataset(self,fftSize):
            dataDict=self.dataDict
            fftSizes=dataDict['fftSizes']
        
            datasetExists=fftSize in fftSizes                
            if datasetExists:
                log.info('use existing dataset. Type={}'.format(self.dataType))
                dataIndex=fftSizes.index(fftSize)
                precomputedDataset=dataDict['precomputedData'][dataIndex]
            else:
                precomputedDataset=self.generateDataset(fftSize)
                self.addDataset(fftSize,precomputedDataset)

            return precomputedDataset
        
        def deleteDatasets(self,fftSizes):
            list(map(self.deleteSingleDataset,np.asarray(fftSizes)))

        def deleteAll(self):
            self.deleteDatasets(self.dataDict['fftSizes'])
        def addDataset(self,fftSize,precomputedDataset):
            dataDict=self.dataDict                
            dataDict['fftSizes'].append(fftSize)
            dataDict['precomputedData'].append(precomputedDataset)


        def generateDataset_MethodFactory(self,*args):
            wavetable_alloc=args[0]
            isNotHalfecomplexType= len(args)==2
            if isNotHalfecomplexType:
                workspace_alloc=args[1]
            
            def generateDataset1(fftSize):
                log.info('generate Dataset for size= {}'.format(fftSize))
                size=c_size_t(fftSize)
                wavetablePointer=wavetable_alloc(size)
                workspacePointer=workspace_alloc(size)
                precomputedDataset=[wavetablePointer,workspacePointer]
                return precomputedDataset
            
            def generateDatasetHalfcomplex(fftSize):
                size=c_size_t(fftSize)
                wavetablePointer=wavetable_alloc(size)
                precomputedDataset=[wavetablePointer]
                return precomputedDataset
            
            if isNotHalfecomplexType:
                generateDataset=generateDataset1
            else:
                generateDataset=generateDatasetHalfcomplex
                
            return generateDataset

        def deleteDataset_MethodFactory(self,*args):
            wavetable_free=args[0]
            isNotHalfecomplexType= len(args)==2
            if isNotHalfecomplexType:
                workspace_free=args[1]
            
            def deleteDataset1(fftSize):
                dataDict=self.dataDict
                fftSizes=dataDict['fftSizes']
                datasetExists=fftSize in fftSizes
                
                if datasetExists:                        
                    dataIndex=fftSizes.index(fftSize)
                    precomputedData=dataDict['precomputedData']
                    precomputedDataset=precomputedData[dataIndex]
                    
                    #free allocated memory
                    wavetablePointer=precomputedDataset[0]
                    workspacePointer=precomputedDataset[1]
                    wavetable_free(wavetablePointer)
                    workspace_free(workspacePointer)

                    #delete dataDict entries
                    del fftSizes[dataIndex]
                    del precomputedData[dataIndex]                        
                else:
                    log.warning('Dataset to Fourier Size {} does not exist! Nothing was deleted.'.format(fftSize))
                    
            def deleteDatasetHalfcomplex(fftSize):
                dataDict=self.dataDict
                fftSizes=dataDict['fftSizes']
                datasetExists=fftSize in fftSizes
                
                if datasetExists:                        
                    dataIndex=fftSizes.index(fftSize)
                    precomputedData=dataDict['precomputedData']
                    precomputedDataset=precomputedData[dataIndex]
                    
                    #free allocated memory
                    wavetablePointer=precomputedDataset[0]
                    wavetable_free(wavetablePointer)

                    #delete dataDict entries
                    del fftSizes[dataIndex]
                    del precomputedData[dataIndex]                        
                else:
                    log.warning('Dataset to Fourier Size {} does not exist! Nothing was deleted.'.format(fftSize))
                    
            if isNotHalfecomplexType:
                deleteDataset=deleteDataset1                                    
            else:
                deleteDataset=deleteDatasetHalfcomplex
            return deleteDataset
            
    dataTypes=FFTPrecomputedData.dataTypes
    def __init__(self):
        dataTypes=self.dataTypes
        FFTPrecomputedData=self.FFTPrecomputedData
        precomputedDataDict={}
        precomputedDataDict[dataTypes['complex']]=FFTPrecomputedData(dataTypes['complex'])
        precomputedDataDict[dataTypes['real']]=FFTPrecomputedData(dataTypes['real'])
        precomputedDataDict[dataTypes['halfcomplex']]=FFTPrecomputedData(dataTypes['halfcomplex'])

        self.precomputedDataDict=precomputedDataDict

        
    def getDataset(self,fftSize,dataType):
        dataTypes=self.dataTypes
        self.dataTypeIsValid(dataType)
        
        dataSet=self.precomputedDataDict[dataType].getDataset(fftSize)

        if dataType==dataTypes['halfcomplex']:
            workspacePointer=self.precomputedDataDict[dataTypes['real']].getDataset(fftSize)[1]
            dataSet=[dataSet[0],workspacePointer]
            
        return dataSet
    
    def deleteDatasets(self,fftSizes,dataType):
        self.dataTypeIsValid(dataType)
        self.precomputedDataDict[dataType].deleteDatasets(fftSizes)
        
    def deleteAll(self,dataType=False):
        if isinstance(dataType,bool):
            for key,value in self.precomputedDataDict.items():
                value.deleteAll()
        else:
            self.dataTypeIsValid(dataType)
            self.precomputedDataDict[dataType].deleteAll()

    def dataTypeIsValid(self,dataType):
        try:
            if dataType in self.dataTypes.values():
                isValidType=True
            else:
                raise AssertionError('FFTPrecomputedData does not know the data type = {} \n Known dataTypes are listed int the class varibale dataTypes'.format(dataType))
        except AssertionError as error:
            log.error(error)
            raise
        return isValidType

class DHTPrecomputedData():
    def __init__(self):
        self.dataDict={}

    def generateDataset(self,dht_size,nus,maxR):
        xmax=c_double(maxR)
        precomputed_data_tuple=()
        for nu in nus:                
            precomputed_data_tuple+=(gsl.gsl_dht_new(c_size_t(dht_size),c_double(abs(nu)),xmax),)
        precomputed_data_array_type=C.POINTER(gslStruct.gsl_dht_struct)*len(nus)
        precomputed_data_array=precomputed_data_array_type(*precomputed_data_tuple)
        precomputed_data_pointer=C.cast(C.pointer(precomputed_data_array),C.POINTER(C.POINTER(gslStruct.gsl_dht_struct)))
        
        log.info('constructed dht r_points ={}, number of orders={} and maxR={}'.format(dht_size+1,len(nus),xmax))
        self.dataDict[(dht_size,nus.tobytes(),maxR)]=precomputed_data_pointer
        return precomputed_data_pointer
    
    def generate_single_dataset(self,dht_size,nu,maxR):
        xmax=c_double(maxR)
        precomputed_data_pointer=gsl.gsl_dht_new(c_size_t(dht_size),c_double(abs(nu)),xmax)
        return precomputed_data_pointer

    def getDataset(self,specifier):
        # specifier is of the form (dht_size,nus,maxR)[float,ndarray or float,float]
        nus=specifier[1]
        #log.info('nus={}'.format(nus))
        if not isinstance(nus,np.ndarray):
            nus=np.array([nus])
            nus.flags.writeable=False
        elif nus.flags.writeable:
            log.warning('Array of nus/orders needs to be readonly! Changing array readability and continue.')
            nus.flags.writeable=False
        specifier=(specifier[0],nus,specifier[2])
        precomputed_dataset=self.dataDict.get((specifier[0],nus.tobytes(),specifier[2]),False)
        dataset_does_not_exists=isinstance(precomputed_dataset,bool)
        if dataset_does_not_exists:
            precomputed_dataset=self.generateDataset(*specifier)
        return precomputed_dataset
    
    def get_single_dataset(self,specifier):
        return self.generate_single_dataset(*specifier)
        
    def deleteAll():
        dataDict=self.dataDict
        list(map(gsl.gsl_dht_free,dataDict.values()))
        dataDict={}

        
class gslMath(GSLInterface):
    def __init__(self):
        gsl.gsl_set_error_handler_off()
        self.dht_precomputed_data_handler=DHTPrecomputedData()
        self.fft_precomputed_data_handler=FFTPrecomputedDataHandler()
        self.fft_data_types=self.fft_precomputed_data_handler.dataTypes

    
    #FFT    
    def assembleFFTArguments(self,data,dataType):
        #I am assuming data is a np array of the proper datatype (np.double for real and halfcomplex and np.complex for complex fft)
        fft_data_types=self.fft_data_types        
        fftSize=data.shape[-1]
        data=data.reshape(-1,fftSize)
        if dataType==fft_data_types['complex']:
            fftSize=int(fftSize/2)
#        log.info('fft size={}'.format(fftSize))
        stride=c_size_t(1)            
        fftSizeC=c_size_t(fftSize)
        precomputedData=self.fft_precomputed_data_handler.getDataset(fftSize,dataType)
  #      log.info('{} arguments =\n{}'.format(dataType,data))
        data_pointers=[data_part.ctypes.data_as(C.POINTER(c_double)) for data_part in data]
                    
        arguments=[data_pointers,repeat(stride),repeat(fftSizeC),repeat(precomputedData[0]),repeat(precomputedData[1])]
        return arguments

    
    def fft_complex_mixedRadix_forward(self,data):
        data=self.fft_complex_pack(data)
        shape=copy(data.shape)
#        log.info('fft complex mixedradix forward data shape ={}'.format(data.shape))
        arguments=self.assembleFFTArguments(data,self.fft_data_types['complex'])
#        log.info('data in forward =\n{}'.format(arguments))
        retVal=map(gsl.gsl_fft_complex_forward,arguments[0],*arguments[1:])
        #        log.info('data out forward =\n{}'.format(data))
        retVal=retVal.reshape(shape)
        unpackedData=self.fft_complex_unpack(data)
#        log.info('data out forward =\n{}'.format(unpackedData))
        return unpackedData
    
    def fft_complex_mixedRadix_inverse(self,data):
        
        data=self.fft_complex_pack(data)
#        log.info('data in inverse =\n{}'.format(data))
        arguments=self.assembleFFTArguments(data,self.fft_data_types['complex'])
        retVal=gsl.gsl_fft_complex_inverse(*arguments)
#        log.info('retVal={}'.format(retVal))
#        log.info('data out inverse =\n{}'.format(data))
        unpackedData=self.fft_complex_unpack(data)
        return unpackedData

    
    def fft_real_mixedRadix_forward(self,data):
        arguments=self.assembleFFTArguments(data,self.fftDataTypes['real'])        
        retVal=gsl.gsl_fft_real_forward(*arguments)
        data=self.fft_halfcomplex_unpack(data)
        return data
    
    
    def fft_halfcomplex_mixedRadix_inverse(self,data):
        data=self.fft_halfcomplex_pack(data)
        arguments=self.assembleFFTArguments(data,self.fftDataTypes['halfcomplex'])
        retVal=gsl.gsl_fft_halfcomplex_inverse(*arguments)
#        log.info('data returned=\n{}'.format(data))
        return data

    def fft_complex_pack(self,data):
#        log.info('input for packing dtype={}'.format(data.dtype))        
        shape=data.real.shape
        new_shape=shape[:-1]+(shape[-1]*2,)
        packed_data=np.zeros(new_shape,dtype=float)
        packed_data[...,0::2] = data.real
        packed_data[...,1::2] = data.imag
        
#        log.info('fft packed_data shape={}'.format(packed_data.shape))
        return packed_data
    
    def fft_complex_unpack(self,data):
  #      log.info('start unpacking')
        unpacked_data=data[...,0::2]+1.j*data[...,1::2]
#        log.info('fft unpacked data shape ={}'.format(unpacked_data.shape))
        return unpacked_data

    def fft_halfcomplex_pack(self,complexArray):
        lenData=len(complexArray)   
        packedArray=np.dstack((complexArray.real,complexArray.imag)).flatten()
        halfcomplexArray=np.concatenate((np.array([packedArray[0]]),packedArray[2:lenData+1]))
        #            log.info('halfcomplexArray={}'.format(halfcomplexArray))
        return halfcomplexArray

    def fft_halfcomplex_unpack(self,data):
        dataLength=len(data)
        unpackedArray=np.zeros(dataLength*2,dtype=np.float)
        
        dataPointer=data.ctypes.data_as(C.POINTER(c_double))
        unpackedArrayPointer=unpackedArray.ctypes.data_as(C.POINTER(c_double))

        stride=c_size_t(1)
        size=c_size_t(dataLength)
        
        gsl.gsl_fft_halfcomplex_unpack(dataPointer,unpackedArrayPointer,stride,size)
        complexArray=unpackedArray[0::2]+1.j*unpackedArray[1::2]
        return complexArray


    #bessel Zero
    
    def besselZero(self,order,zeroNumber):
        order=c_double(abs(order))
        zeroNumber=c_int(zeroNumber)
        zero=gsl.gsl_sf_bessel_zero_Jnu(order,zeroNumber)
        return zero

    #Legendre Polynomial
        
    def legendrePoly(self,degree,x):
        degree=c_int(degree)
        x=c_double(x)
        value=gsl.gsl_sf_legendre_Pl(degree,x)
        return value
        
    #DHT
    def generate_dht_2D(self,n_radial_points,orders,maxR):
        #assumes data is complex numpy array
        #gsl is quite confusing here !! the dhtSize to initialize the transform is M-1 where M indicates the number of sample points where f(u_M) is supposed to be 0
        dht_size=n_radial_points
        n_orders=len(orders)
        c_n_orders=c_size_t(n_orders)
        data_specifier=(dht_size,orders,maxR)
        precomputed_data_pointer=self.dht_precomputed_data_handler.getDataset(data_specifier)

        reciprocal_coeff=np.zeros((n_orders,n_radial_points),dtype=np.complex,order='C')
        coeff=np.zeros((n_orders,n_radial_points),dtype=np.complex,order='C')

        #forward_prefactor=1.j**(-1*orders)*4*np.pi*maxR**2/np.array(tuple(map(self.besselZero,orders,np.full(n_orders,dht_size+1))))
        #inverse_prefactor=1.j**orders/(np.pi*maxR**2)

        forward_prefactor=((-1.j)**orders)*4*np.pi*(maxR/np.array(tuple(map(self.besselZero,orders,np.full(n_orders,dht_size+1)))))**2
        inverse_prefactor=(1.j**orders)/(np.pi*maxR**2)
        
        def dht_2D_forward(harmonic_coeffs):
            gsl.dht_2D(c_n_orders,precomputed_data_pointer,harmonic_coeffs,reciprocal_coeff,forward_prefactor)
            return reciprocal_coeff
        
        def dht_2D_inverse(reciprocal_harmonic_coeffs):
            gsl.dht_2D(c_n_orders,precomputed_data_pointer,reciprocal_harmonic_coeffs,coeff,inverse_prefactor)
            return coeff
        return dht_2D_forward,dht_2D_inverse
        
    def generate_dht_bare(self,n_radial_points,order,maxR,prefactor):
        dht_size=n_radial_points
        data_specifier=(dht_size,order,maxR)
        precomputed_data_pointer=self.dht_precomputed_data_handler.getDataset(data_specifier)

        complex2=c_double*2
        prefactor=gslStruct.gsl_complex(complex2(prefactor.real,prefactor.imag))
        
        coeff=np.zeros(n_radial_points,dtype=np.complex,order='C')

        def dht_bare(harmonic_coeff):
            gsl.dht_bare(precomputed_data_pointer,harmonic_coeff,coeff,prefactor)
            return coeff
        return dht_bare

    def dht_test(self,data):
        gsl.dht_test(c_size_t(data.shape[0]),c_size_t(data.shape[1]),data)
    
        
    
    def dht_forward2(self,data,nu,maxR):        
        #assumes data is complex numpy array
        #gsl is quite confusing here !! the dhtSize to initialize the transform is M-1 where M indicates the number of sample points where f(u_M) is supposed to be 0
        dhtSize=len(data)


        transformedRealData=np.zeros(dhtSize,dtype=np.float)
        transformedImagData=np.zeros(dhtSize,dtype=np.float)
        dataReal=np.copy(data.real)
        dataImag=np.copy(data.imag)

        initializedWorkspacePointer=self.dht_precomputed_data_handler.get_single_dataset((dhtSize,nu,maxR))
        gsl.gsl_dht_apply(initializedWorkspacePointer,dataReal.ctypes.data_as(C.POINTER(c_double)),transformedRealData.ctypes.data_as(C.POINTER(c_double)))
        gsl.gsl_dht_apply(initializedWorkspacePointer,dataImag.ctypes.data_as(C.POINTER(c_double)),transformedImagData.ctypes.data_as(C.POINTER(c_double)))
        transformedData=transformedRealData+1.j*transformedImagData
        return transformedData

    def dht_inverse2(self,data,nu):
        #not precisely invertible for low values of M = len(data)+1
        dhtSize=len(data)
        besselZero=self.besselZero(nu,dhtSize+1)
        transformedData=(besselZero**2)*self.dht_forward(data,nu)
        return transformedData





import pytest
import os
import sys
import subprocess
import numpy as np
import importlib
import numpy as np
from tlib import copy,save,load,create_path_if_nonexistent
import xframe
config_path = os.path.expanduser('~/.xframe/config.py')
config_backup_path = os.path.expanduser('~/.xframe/backup_config.py')

tmp_config_path = ''
_path = ''

'''
Disclaimer: This module houses integration tests for the fxs module.
Currently this means that the functions:

correlate
simulate_ccd
extract
reconstruct
average

will be executed in sequence for sample structure with small grid size.
After each execution we check whether the output files exist and contain roughly sensible data.
No detailed checks on the output data is performed.
'''

@pytest.fixture(scope='session')
def tmp_path(tmpdir_factory):
    path = tmpdir_factory.mktemp('xtest'+str(np.random.rand()))
    return path

@pytest.fixture(scope='session',autouse=True)
def set_temp_home(tmp_path):    
    config_file_exists = os.path.exists(config_path)
    global tmp_config_path
    tmp_config_path = f'{tmp_path}/.xframe/'
    if config_file_exists:
        initial_config = load(config_path)
        print(f'initial config content = {initial_config}')
        copy(config_path,config_backup_path)
    xframe.change_home(tmp_config_path)
    
    yield
    if config_file_exists:
        copy(config_backup_path,config_path)
    else:
        os.remove(config_file)
    print(f'config after test = {load(config_path)}')
    
def home_files_exist():
    paths=[
        'projects/fxs',
        'projects/tutorial',
        'experiments/SPB',
        'settings/projects/fxs/correlate/tutorial.yaml',
        'settings/projects/fxs/correlate/default_0.01.yaml',
        'settings/projects/fxs/simulate_ccd/tutorial.yaml',
        'settings/projects/fxs/simulate_ccd/default_0.01.yaml',
        'settings/projects/fxs/extract/tutorial.yaml',
        'settings/projects/fxs/extract/default_0.01.yaml',
        'settings/projects/fxs/reconstruct/tutorial.yaml',
        'settings/projects/fxs/reconstruct/default_0.01.yaml',
        'settings/projects/fxs/average/tutorial.yaml',
        'settings/projects/fxs/average/default_0.01.yaml',
        'settings/experiments/SPB/tutorial.yaml',
        'settings/experiments/SPB/default_0.01.yaml',
    ]
    global tmp_config_path
    for path in paths:
        _path= os.path.join(tmp_config_path,path)
        exists = os.path.exists(_path)
        if not exists:
            assert _path == None


def check_settings(file_path):
    '''
    Checks if file contents are same as converted(replacing numpy arrays and types)
    '''
    loaded_settings = xframe.database.default.load(file_path)
    converter = xframe.settings.tools.SettingsConverter
    program_settings = converter.convert(xframe.settings.project.dict())
    raw_settings = converter.convert(xframe.settings.raw_project)
    assert (loaded_settings == program_settings) or (loaded_settings == raw_settings)
def check_dict_content(_dict,template):
    for key,temp in template.items():
        assert key in _dict, f'Key "{key}" could not be found.'
        val = _dict[key]
        if isinstance(temp,dict):
            check_dict_content(val,temp)
        elif isinstance(temp,(tuple,list)):
            assert isinstance(val,np.ndarray),key+' : '+str(_dict.keys())
            assert val.dtype == temp[0],key+' : '+str(_dict.keys())
            assert val.shape == temp[1],key+' : '+str(_dict.keys())
            assert np.isnan(val).any() == False,key+' : '+str(_dict.keys())
        else:
            assert isinstance(val,temp),key+' : '+str(_dict.keys())
    

@pytest.fixture(scope='session')
def patterns():
    patterns = np.random.rand(*TestsCorrelate.pattern_shape).astype(np.float32)
    paths = [os.path.join(tmp_config_path,TestsCorrelate.pattern_folder,f'{i}.raw') for i in range(TestsCorrelate.pattern_shape[0])]
    for p,d in zip(paths,patterns):
        xframe.database.default.save(p,d)
    pattern_list = [p+'\n' for p in paths]
    xframe.database.default.save(os.path.join(tmp_config_path,TestsCorrelate.pattern_list_path),pattern_list)
      
@pytest.fixture(scope='session')
def run_correlate(patterns):
    pattern_shape = TestsCorrelate.pattern_shape    
    settings_path= os.path.join(tmp_config_path,'settings/projects/fxs/correlate/test.yaml')
    settings = [
        "structure_name: '3d_test_corr'\n",
        "image_dimensions:\n",
        f"  command: '{pattern_shape[1:]}'\n",
        "n_processes: 32\n",
        f"max_n_patterns: {pattern_shape[0]}\n",
        f"detector_origin: {[(pattern_shape[1]-1)/2]*2}\n",
        "phi_range:\n",
        f"  command: '(0,2*np.pi,{pattern_shape[1]*2},\"exact\")'\n",
        f"fc_n_max: {pattern_shape[1]-1}\n",
        "IO:\n",
        "  folders:\n",
        "    in_base:\n",
        "      data: 'correlate/'\n"
    ]
    save(settings_path,settings)   
    xframe.select_project('fxs','correlate','test')
    xframe.run()
        
class TestsCorrelate:
    pattern_shape = (200,16,16)
    pattern_folder = 'data/fxs/correlate/patterns/'
    pattern_list_path = 'data/fxs/correlate/patterns_list.txt'
    def test_files_exist(self,run_correlate):
        date=xframe.database.project.get_time_string()
        files = [
            'data/fxs/ccd/3d_test_corr.h5',
            f'data/fxs/ccd/archive/3d_test_corr/{date}/run_0/settings.yaml',
            f'data/fxs/ccd/archive/3d_test_corr/{date}/run_0/ccd.h5',
        ]
        for fpath in files:
            path= os.path.join(tmp_config_path,fpath)
            exists = os.path.exists(path)
            assert exists,f'{path} does not exist'
    def test_file_contents(self,run_correlate):
        date=xframe.database.project.get_time_string()
        settings_path = os.path.join(tmp_config_path,f'data/fxs/ccd/archive/3d_test_corr/{date}/run_0/settings.yaml')
        check_settings(settings_path)
        self.check_h5_file_contents()
    def check_h5_file_contents(self):
        date=xframe.database.project.get_time_string()
        h5_path = os.path.join(tmp_config_path,f'data/fxs/ccd/archive/3d_test_corr/{date}/run_0/ccd.h5')
        data = xframe.database.default.load(h5_path)
        ccd_shape=(self.pattern_shape[1]//2,)*2+(self.pattern_shape[1]*2,)        
        data_structure = {'radial_points':[float,(self.pattern_shape[1]//2,)],
                          'angular_points':[float,(self.pattern_shape[1]*2,)],
                          'xray_wavelength':float,
                          'average_intensity':[float,(self.pattern_shape[1]//2,)],
                          'cross_correlation':{
                              'I1I1':[float,ccd_shape]
                          }
                          }
        check_dict_content(data,data_structure)


@pytest.fixture(scope='session')
def run_simulate_ccd():
    radial_points = TestsCorrelate.pattern_shape[1]//2
    max_order = TestsCorrelate.pattern_shape[1]
    settings_path= os.path.join(tmp_config_path,'settings/projects/fxs/simulate_ccd/test.yaml')
    settings = [
        "structure_name: '3d_test'\n",
        "dimensions: 3\n",
        "grid:\n",
        "  max_q: 0.017\n",
        f"  n_radial_points: {radial_points}\n",
        f"  max_order: {max_order}\n",
        "shapes:\n",
        "  types:\n",
        "    command: \"['sphere']*2\"\n",
        "  centers:\n",
        "    command: \"[(140.0,np.pi/2,phi*2*np.pi/2) for phi in range(2) ]\"\n",
        "  sizes:\n",
        "    command: \"[70]*2\"\n",
        "  densities: [25,50]\n",
        "cross_correlation:\n",
        "  method: 'back_substitution'\n",
        "  xray_wavelength: 1.23984\n",
    ]
    save(settings_path,settings)   
    xframe.select_project('fxs','simulate_ccd','test')
    xframe.run()
    
class TestsSimulateCCD:    
    def test_files_exist(self,run_simulate_ccd):
        date=xframe.database.project.get_time_string()
        files = [
            'data/fxs/ccd/3d_test.h5',
            f'data/fxs/ccd/archive/3d_test/{date}/run_0/settings.yaml',
            f'data/fxs/ccd/archive/3d_test/{date}/run_0/ccd.h5',
            f'data/fxs/ccd/archive/3d_test/{date}/run_0/model_density.vts',
        ]
        for fpath in files:
            path= os.path.join(tmp_config_path,fpath)
            exists = os.path.exists(path)
            assert exists,f'{path} does not exist'
    def test_files_contents(self,run_simulate_ccd):
        date=xframe.database.project.get_time_string()
        settings_path = os.path.join(tmp_config_path,f'data/fxs/ccd/archive/3d_test/{date}/run_0/settings.yaml')
        check_settings(settings_path)
        self.check_h5_file_contents()
    def check_h5_file_contents(self):
        date=xframe.database.project.get_time_string()
        h5_path = os.path.join(tmp_config_path,f'data/fxs/ccd/archive/3d_test/{date}/run_0/ccd.h5')
        data = xframe.database.default.load(h5_path)
        pattern_shape = TestsCorrelate.pattern_shape
        n_angular_points = 2*pattern_shape[1] 
        ccd_shape=(pattern_shape[1]//2,)*2+(n_angular_points,)
        data_structure = {'radial_points':[float,(pattern_shape[1]//2,)],
                          'angular_points':[float,(n_angular_points,)],
                          'xray_wavelength':float,
                          'average_intensity':[float,(pattern_shape[1]//2,)],
                          'cross_correlation':{
                              'I1I1':[float,ccd_shape]
                          }
                          }
        check_dict_content(data,data_structure)    

    
@pytest.fixture(scope='session')
def run_extract(run_simulate_ccd):
    radial_points = TestsCorrelate.pattern_shape[1]//2
    max_order = TestsCorrelate.pattern_shape[1]-1
    settings_path= os.path.join(tmp_config_path,'settings/projects/fxs/extract/test.yaml')
    settings = [
        "structure_name: 'test'\n",
        "dimensions: 3\n",
        f"max_order: {max_order}\n",
        "IO:\n",
        "  files:\n"
        "    ccd:\n",
        "      name: 3d_test.h5\n",
        "    invariants:\n",
        "      options:\n",
        "        plot_first_invariants: False\n",
        "        plot_first_invariants_from_proj_matrices: False\n",
    ]
    save(settings_path,settings)   
    xframe.select_project('fxs','extract','test')
    xframe.run()    
class TestExtract:
    def test_files_exist(self,run_extract):
        date=xframe.database.project.get_time_string()
        files = [
            'data/fxs/invariants/3d_test.h5',
            f'data/fxs/invariants/archive/3d_test/{date}/run_0/extraction_settings.yaml',
            f'data/fxs/invariants/archive/3d_test/{date}/run_0/proj_data.h5',
        ]
        for fpath in files:
            path= os.path.join(tmp_config_path,fpath)
            exists = os.path.exists(path)
            assert exists,f'{path} does not exist'
    def test_files_contents(self,run_extract):
        date=xframe.database.project.get_time_string()
        settings_path = os.path.join(tmp_config_path,f'data/fxs/invariants/archive/3d_test/{date}/run_0/extraction_settings.yaml')
        check_settings(settings_path)
        self.check_h5_file_contents()
     def check_h5_file_contents(self):
        date=xframe.database.project.get_time_string()
        h5_path = os.path.join(tmp_config_path,f'data/fxs/ccd/archive/3d_test/{date}/run_0/ccd.h5')
        data = xframe.database.default.load(h5_path)
        pattern_shape = TestsCorrelate.pattern_shape
        n_angular_points = 2*pattern_shape[1]
        max_order = pattern_shape[1]
        n_radial_points = pattern_shape[0] 
        ccd_shape=(pattern_shape[1]//2,)*2+(n_angular_points,)
        bl_shape=(max_order,)+(n_radial_points,)*2
        data_structure = {'dimensions':int,
                          'xray_wavelength':float,
                          'max_order':float,
                          'average_intensity',[float,n_radial_points],
                          'data_angular_points':[float,(n_angular_points,)],
                          'data_radial_points':[float,(n_radial_points,)],
                          'deg_2_invariant':{'I1I1':[complex,bl_shape]}
                          'deg_2_invariant_masks':{'I1I1':[bool,bl_shape]}
                          }
                          }
        check_dict_content(data,data_structure)    
        pr_matrices=data['data_projection_matrices']['I1I1']
        for l,mat in enumerate(pr_matrices):
            assert mat.dtype== complex, f'projection matrix of order {l} is not of dtype complex but {mat.dtype}.'
            assert np.isnan(mat).any() == False, f'projection matrix of order {l} contains Nans.'

            mat_shape = (n_radial_points,min(2*l+1,n_radial_points))
            assert mat.shape == mat_shape ,f'projection matrix of order {l} has wring shape. {mat.shape} instead of {mat_shape}'

        
@pytest.fixture(scope='session')
def run_reconstruct(run_extract):
    radial_points = TestsCorrelate.pattern_shape[1]//2
    max_order = TestsCorrelate.pattern_shape[1]-1
    settings_path= os.path.join(tmp_config_path,'settings/projects/fxs/reconstruct/test.yaml')
    settings = [
        "structure_name: 'test'\n",
        "dimensions: 3\n",
        f"particle_radius: 250\n",
        "grid:\n",
        "  n_radial_points: 8\n",
        "  max_order: 15\n",
        "projections:\n",
        "  reciprocal:\n",
        "    use_order_ids:\n",
        "      command: 'np.arange(16)'\n",
        "multi_process:\n",
        "  n_parallel_reconstructions: 10\n",
        "main_loop:\n",
        "  sub_loops:\n",
        "    main:\n",
        "      iterations: 2\n",
        "IO:\n",
        "  files:\n",
        "    reconstructions:\n",
        "      options:\n",
        "        plot_first_used_invariants: False\n",
        "        plot_reconstructed_deg2_invariants: False\n"
    ]
    save(settings_path,settings)   
    xframe.select_project('fxs','reconstruct','test')
    xframe.run()
    
class TestReconstruct:
    def test_files_exist(self,run_reconstruct):
        date=xframe.database.project.get_time_string()
        files = [
            f'data/fxs/reconstructions/3d_test/{date}/run_0/settings.yaml',
            f'data/fxs/reconstructions/3d_test/{date}/run_0/data.h5',
            f'data/fxs/reconstructions/3d_test/{date}/run_0/vtk/',
        ]
        for fpath in files:
            path= os.path.join(tmp_config_path,fpath)
            exists = os.path.exists(path)
            assert exists,f'{path} does not exist'        
    def test_files_contents(self,run_reconstruct):
        date=xframe.database.project.get_time_string()
        settings_path = os.path.join(tmp_config_path,f'data/fxs/reconstructions/3d_test/{date}/run_0/settings.yaml')
        check_settings(settings_path)

    
@pytest.fixture(scope='session')
def run_average(run_reconstruct):
    settings_path= os.path.join(tmp_config_path,'settings/projects/fxs/average/test.yaml')
    settings = [
        "structure_name: 'test'\n",
        "reconstruction_files:\n",
        "  - 3d_test/{today}/run_0/data.h5\n",        
        "multi_process:\n",
        "  use: True\n",
        "  n_processes: 2\n",
        "find_rotation:\n",
        "  r_limit_ids:\n",
        "    command: 'np.arange(4)'\n",        
    ]
    save(settings_path,settings)   
    xframe.select_project('fxs','average','test')
    xframe.run()
    
class TestAverage:
    def test_files_exist(self,run_average):
        date=xframe.database.project.get_time_string()
        files = [
            f'data/fxs/averages/3d_test/{date}/run_0/settings.yaml',
            f'data/fxs/averages/3d_test/{date}/run_0/average_results.h5',
            f'data/fxs/averages/3d_test/{date}/run_0/PRTF.png',
            f'data/fxs/averages/3d_test/{date}/run_0/vtk/real_average.vts',
            f'data/fxs/averages/3d_test/{date}/run_0/vtk/reciprocal_average.vts',
        ]
        for fpath in files:
            path= os.path.join(tmp_config_path,fpath)
            exists = os.path.exists(path)
            assert exists,f'{path} does not exist'
    def test_files_contents(self,run_average):
        date=xframe.database.project.get_time_string()
        settings_path = os.path.join(tmp_config_path,f'data/fxs/averages/3d_test/{date}/run_0/settings.yaml')
        check_settings(settings_path)

